! This file is part of FD4: Four-Dimensional Distributed Dynamic Data structures
!  
! Copyright 2010-2011 Matthias Lieber, ZIH, TU Dresden, Germany
! matthias.lieber@tu-dresden.de
!  
! FD4 is free software: you can redistribute it and/or modify
! it under the terms of the GNU General Public License as published by
! the Free Software Foundation, either version 3 of the License, or
! (at your option) any later version.
!  
! FD4 is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
! GNU General Public License for more details.
!  
! You should have received a copy of the GNU General Public License
! along with FD4.  If not, see <http://www.gnu.org/licenses/>.

!--------------------------------------------------------------!
!                                                              !
!            Coupling of FD4 to other parallel model           !
!                                                              !
!--------------------------------------------------------------!

#include "fd4.h"

!! Coupling of FD4 to other parallel model.
!!
!! *Limitations*:
!!  - sequential coupling only (i.e. FD4 operates on same processes
!!    as the coupled model)
!!  - user must ensure that couple arrays do not overlap, otherwise same data
!!    will be transferred multiple times. Problematic when using $fd4_couple_put!
!!
!! Preprocessor options:
!!  - $FD4_COUPLE_DEBUG: verbose debug output to fort.* files
!!  - $FD4_COUPLE_CHECK_RECV_SIZE: check if received message size equals expected message size, changes from
!!    non-blocking receive to blocking receive, aborts on error
!!
!! Example:
!!> real(r_k), allocatable :: array_Rho(:,:,:,:), array_Tmp(:,:,:,:)
!!  real(r_k), pointer :: arraypointer(:,:,:)
!!  type(fd4_couple) :: couple
!!
!!  [...]
!!  initialize(array_Rho, array_Tmp, ...)
!!  [...]
!!
!!  ! Create the couple context.
!!  call fd4_couple_create(couple, domain, err)
!!
!!  ! Loop over all MPI processes to tell FD4 coupling metadata.
!!  do irank=0,nproc-1
!!    ! Set cabnd(3,2) to the bounds of the coupling arrays of current rank.
!!    ! (here we simply calculate bounds for an 1D-partitioning)
!!    cabnd(1:3,1) = (/   1+(irank*sizeX)/nproc,     1 ,     1 /)
!!    cabnd(1:3,2) = (/ ((irank+1)*sizeX)/nproc, sizeY , sizeZ /)
!!    ! Add a couple array of current rank with given bounds. Associate it with variable
!!    ! index var_Rho and the given time step index. FD4 assigns an index to local couple
!!    ! arrays (i.e. those that belong to the calling rank) to caidx_Rho.
!!    call fd4_couple_add_array(couple, irank, var_Rho, step, cabnd, caidx_Rho, err)
!!    ! Add another couple array of current rank.
!!    call fd4_couple_add_array(couple, irank, var_Tmp, step, cabnd, caidx_Tmp, err)
!!  end do
!!
!!  ! Assign the pointer and offset to relevant data of the local couple array.
!!  ! The argument must be a pointer. Use an additional pointer to point to an
!!  ! allocatable where required (contiguous subsections are allowed).
!!  arraypointer => array_Rho(:,:,:,2)
!!  call fd4_couple_set_local_3D_array(couple, caidx_Rho, arraypointer, (/1,1,1/), err)
!!  arraypointer => array_Tmp(:,:,:,1)
!!  call fd4_couple_set_local_3D_array(couple, caidx_Tmp, arraypointer, (/1,1,1/), err)
!!
!!  ! Commit the couple context, FD4 now checks the couple arrays and prepares MPI data types
!!  call fd4_couple_commit(couple, err)
!!
!!  ! Finally put data to FD4's data structures
!!  call fd4_couple_put(couple, err)
!!<
!!
!! Some notes about sorting of the components of the mpi datatypes we are building here:
!!  - components are identified by position and variable/step
!!  - sorting is required for correct matching on sender and receiver side
!!  - sorting happens only for the position (in $fd4_mpitype_commit)
!!  - variable/step ordering is given by the order in the array description $desc
!!  - that means
!!    - sorting in $fd4_mpitype_commit has to be stable!
!!    - *all processes put the arrays in the same order into* $desc
!!
!! *TODO*:
!!  - do as much as possible of checking, min disp search, etc. which is done in $fd4_couple_get/$put in the set/chg_array routines
!!  - use the same technique as used for the cpltypestore (types for FD4 blocks) for the arraytypes (types for couple arrays)
!!    to prevent duplicates (hide in $fd4_couple_build_cplarray_type)
!!  - implement a routine which checks the couple description for overlapping arrays
!!    (which are not allowed for $fd4_couple_put!)
!!     - this routine should be called in couple_create
!!     - it should disable $fd4_couple_put if necessary
!!
!! @see fd4_domain_mod
!! @author Matthias Lieber
module fd4_couple_mod

  use fd4_globaldef_mod
  use fd4_vartab_mod
  use fd4_block_mod
  use fd4_domain_mod
  use fd4_iter_mod
  use fd4_mpi_mod
  use fd4_mpitype_mod
  use fd4_comm_mod
  use fd4_stat_mod
  use fd4_neigh_mod
  use rbtree_fd4_neigh_mod
  use rbtree_int_mod
  use rbtree_int8_mod
  use stack_mod
  use stack3_mod
  use section_mod

  USE_MPI
  
  implicit none
  private
  
  INCLUDE_MPIFH
  
  
  ! public types
  public :: fd4_couple
  public :: fd4_couple_local_desc
  public :: fd4_domain
  public :: fd4_mpitype
  ! public subroutines
  public :: fd4_couple_create
  public :: fd4_couple_add_array
  public :: fd4_couple_set_local_3D_array
  public :: fd4_couple_set_local_4D_array
  public :: fd4_couple_commit
  public :: fd4_couple_chg_local_3D_array
  public :: fd4_couple_chg_local_4D_array
  public :: fd4_couple_delete
  public :: fd4_couple_put
  public :: fd4_couple_get  
  public :: fd4_couple_mark_blocks
  public :: fd4_couple_debug_metadata
  ! public parameters
  public :: FD4_CPL_GET, FD4_CPL_PUT, FD4_CPL_BOTH


  !! For the last dimension of $arraytypes, the field of pre-built MPI types for couple array sections:
  !! type for $put, type for $get, type for $get without additional face cell (face variables only),
  !! indicator of case for $get (see $CPL_GET_CASE_*) which depends on state of neighbor FD4 blocks
  integer(i_k), parameter :: CPL_PUT=1, CPL_GET=2, CPL_GET_NOFACE=3, CPL_GET_CASE=4

  !! Indicators of cases for each (FD4 block, couple array) pair which make the selection
  !! of the MPI type for face variables in $fd4_couple_get dependent on the state of the
  !! neighbor blocks (neighbor in the direction of the face).
  !! Used in $arraytypes(:,:,:,:,$CPL_GET_CASE).
  !!
  !! case 0: no intersection between couple array and FD4 block
  integer(i_k), parameter :: CPL_GET_CASE_SKIP=0
  !! case A: use $CPL_GET type (no dependence on neighbor blocks or no face variable)
  !! (last cell of couple array resides in FD4 block but is not first cell)
  integer(i_k), parameter :: CPL_GET_CASE_NORMAL=1
  !! case B: if left neighbor exists: do not receive, else: use $CPL_GET type
  !! (last cell of couple array is first cell of FD4 block, left neighbor could provide the data)
  integer(i_k), parameter :: CPL_GET_CASE_CHECK_L=2
  !! case C: if right neighbor exists: use $CPL_GET_NOFACE type, else: use $CPL_GET type
  !! (last cell of couple array is right outside the FD4 block, right neighbor could provide additional face cell)
  integer(i_k), parameter :: CPL_GET_CASE_CHECK_R=3

  !! states of couple context:
  !! new, created, added vars, added local vars, commited
  integer(i_k), parameter :: CPL_NEW=1, CPL_CRT=2, CPL_ADD=3, CPL_LADD=4, CPL_CMMT=5
  
  !! indicators to restrict usage of couple context to fd4_couple_get or fd4_couple_put
  !! to save some memory and preparation work
  integer(i_k), parameter :: FD4_CPL_GET = 1, FD4_CPL_PUT = 2
  integer(i_k), parameter :: FD4_CPL_BOTH = IOR(FD4_CPL_GET, FD4_CPL_PUT)
  
  !! indexes in mpitypes(1:2) of the neigh, one type is used for the FD4 side, the other for the couple array
  integer(i_k), parameter :: FD4_NEIGH_CPL=1, FD4_NEIGH_FD4=2

  !! Description of one *local* coupling array.
  !! Used only internally.
  !!
  !! - 3D data: $array and $offset
  !! - 4D data: $array4, $offset, and $offset4
  !! - $array and $array4 must not be set at the same time
  type fd4_couple_local_desc
    real(r_k), pointer :: array(:,:,:) => NULL()   !! data array
    real(r_k), pointer :: array4(:,:,:,:) => NULL()!! 4-dimensional data array (for variables with bins)
    integer(i_k) :: offset(3) = (/1,1,1/)          !! spatial offset of the data in array, e.g. (/1,1,1/)
    integer(i_k) :: offset4 = 1                    !! offset of the data in the 4th dimension
    integer(MPI_ADDRESS_KIND) :: disp              !! memory address of the array
  end type fd4_couple_local_desc

  
  !! The FD4 couple context.
  !!
  type fd4_couple
    type(fd4_domain), pointer :: domain => NULL()  !! the domain
    integer(i_k) :: loopbnd(3,2)                   !! bounds for the loop over blocks    
    integer, pointer :: arraytypes(:,:,:,:,:) => NULL() !! pre-built MPI types for couple arrays
    type(rbtree_fd4_neigh) :: neighstore           !! MPI types for each communication partner are stored in a rbtree
    integer(i_k) :: ncplarrays = 0                 !! number of local couple arrays in couple description
    integer(i_k), pointer :: rankbnd(:,:,:) => NULL() !! union bounds of all couple arrays for each rank
    integer(i_k) :: faceoffset(3)                  !! offset to local block bounds union due to face variables
    ! new data structure 2010-03-17
    integer :: state = CPL_NEW                     !! state of couple context
    integer, pointer :: desc_rank(:) => NULL()     !! starting points in desc_var per rank 
    type(stack3) :: desc_var                       !! (idx, step, abnd_index) of all couple arrays
    type(stack3) :: desc_abnd                      !! abnd (two stack3 entries) of a couple array
    type(fd4_couple_local_desc), pointer &
                 :: desc_local(:) => NULL()        !! pointer and offset of local coupling arrays
    integer(i_k) :: ins_rank = 0                   !! last inserted rank
    integer(i_k) :: cpldir = FD4_CPL_BOTH          !! direction of coupling (get/put/both)
    integer(i_k) :: stat_put, stat_get             !! index of transferred bytes in the statistics object
    logical :: face_overlap = .true.               !! do couple arrays for face variables overlap?
  end type fd4_couple

  
contains  


  include 'util/stack.incf'
  include 'util/stack3.incf'
  include 'framework/fd4_domain.incf'
  

  !! Print some debug information about coupling meta data to files for each rank
  !!
  subroutine fd4_couple_debug_metadata(cpl, name)
    type(fd4_couple),  intent(in),  target     :: cpl          !! couple context
    character(*),      intent(in)              :: name         !! file name prefix
    
    integer(i_k) :: i, ofile
    type(fd4_vartab), pointer :: var
        
    ofile = 1234
    open(unit=ofile,file=name//'-'//i2s(cpl%domain%mpi%rank,.true.)//'.txt',action='write',position='rewind',status='replace')

    write(ofile,'(A,I6,A)') 'cpl%desc_abnd (',cpl%desc_abnd%pos/2,')'
    do i=1,cpl%desc_abnd%pos,2
      write(ofile,'(I6,A,3(I6),A,3(I6))') i,': ',cpl%desc_abnd%a(1:3,i),'  -  ',cpl%desc_abnd%a(1:3,i+1)
    end do
    
    write(ofile,'(A,I6,A)') 'cpl%desc_var (',cpl%desc_var%pos,')'
    write(ofile,'(A6,A,7(A6))') ' ','  ','idx','st','abnd',' ','bins','face','name'
    do i=1,cpl%desc_var%pos
      var => cpl%domain%vartab( cpl%desc_var%a(1,i) )
      write(ofile,'(I6,A,3(I6),A6,2(I6),A2,A)') i,': ',cpl%desc_var%a(1:3,i),' ',var%nbins,var%facevar,' ',var%name
    end do
    
    if(associated(cpl%desc_rank)) then
      write(ofile,'(A,I6,A)') 'cpl%desc_rank (',size(cpl%desc_rank),')'
      do i=lbound(cpl%desc_rank,1),ubound(cpl%desc_rank,1)
        write(ofile,'(I6,A,I6)') i,': ',cpl%desc_rank(i)
      end do
    end if

    if(associated(cpl%desc_local)) then
      write(ofile,'(A,I6,A)') 'cpl%desc_local (',size(cpl%desc_local),')'
      do i=lbound(cpl%desc_local,1),ubound(cpl%desc_local,1)
        if(associated(cpl%desc_local(i)%array)) then
          write(ofile,'(I6,A,A6,3(I6),A,A6,3(I6),A,A6,3(I6))') i,': ',' ',lbound(cpl%desc_local(i)%array),'  -  ', &
                           ' ',ubound(cpl%desc_local(i)%array),'  @  ',' ',cpl%desc_local(i)%offset
        elseif(associated(cpl%desc_local(i)%array4)) then
          write(ofile,'(I6,A,4(I6),A,4(I6),A,4(I6))') i,': ',lbound(cpl%desc_local(i)%array4),'  -  ', &
                             ubound(cpl%desc_local(i)%array4),'  @  ',cpl%desc_local(i)%offset4,cpl%desc_local(i)%offset
        else
          write(ofile,'(I6,A,A24,A,A24,A,4(I6))') i,': ',' (not associated) ','  -  ', &
                             ' ','  @  ',cpl%desc_local(i)%offset4,cpl%desc_local(i)%offset
        end if        
      end do
    end if
    
    close(ofile)
  
  end subroutine fd4_couple_debug_metadata
  
  
  !! Create FD4 couple context.
  !!
  !! User may pass hints for the meta data creation ($opt_nvar, $opt_nabnd).
  !! Giving accurate hints speeds up the process of adding coupling arrays and reduces memory consumption.
  !! If possible, make use of these hints!
  !!
  !! With $opt_cpldir you can restrict the use of the couple context to $fd4_couple_get
  !! and $fd4_couple_put/$fd4_mark_blocks.
  !! Use the parameters $FD4_CPL_GET and $FD4_CPL_PUT for these purpose.
  !! The default is no restriction ($FD4_CPL_BOTH).
  !! The restriction is used by $fd4_couple to reduce the effort in preparing MPI data types.
  !!
  subroutine fd4_couple_create(cpl, domain, err, opt_nvar, opt_nabnd, opt_cpldir)
    type(fd4_couple),     intent(inout)        :: cpl          !! couple context
    type(fd4_domain),     intent(inout),target :: domain       !! the domain
    integer(i_k),optional,intent(in)           :: opt_nvar     !! number of coupling arrays per rank (hint)
    integer(i_k),optional,intent(in)           :: opt_nabnd    !! number of different global coupling array bounds per rank (hint)
    integer(i_k),optional,intent(in)           :: opt_cpldir   !! restriction of use (get/put/both)
    integer(i_k),         intent(out)          :: err          !! error status: 0...ok
  
    err = -1
    FD4_L1_ASSERT(cpl%state/=CPL_NEW,'cpl not new')
    
    cpl%domain => domain
    
    allocate(cpl%desc_rank(0:domain%mpi%nprocs))
    cpl%desc_rank = -1
    
    if(present(opt_nvar)) then
      call stack3_create(cpl%desc_var,domain%mpi%nprocs*opt_nvar)
    else
      call stack3_create(cpl%desc_var,domain%mpi%nprocs)
    end if

    if(present(opt_nabnd)) then
      call stack3_create(cpl%desc_abnd,domain%mpi%nprocs*2*opt_nabnd)
    else
      call stack3_create(cpl%desc_abnd,domain%mpi%nprocs)
    end if
    
    if(present(opt_cpldir)) then
      cpl%cpldir = opt_cpldir
    else
      cpl%cpldir = FD4_CPL_BOTH
    end if
    
    cpl%ins_rank = -1
    cpl%state = CPL_CRT
    cpl%ncplarrays = 0
    
    call fd4_stat_add(domain%stat,'Couple put Bytes',cpl%stat_put)
    call fd4_stat_add(domain%stat,'Couple get Bytes',cpl%stat_get)
    
    err = 0

  end subroutine fd4_couple_create


  
  !! Add a coupling array to the FD4 couple context.
  !!
  !! Only allowed after $fd4_couple_create and before $fd4_couple_set_local_*D_array.
  !! Consecutive calls must have equal or increasing $rank arguments.
  !! *All* ranks must call this function for *all* coupling arrays (own and all other ranks').
  !!
  subroutine fd4_couple_add_array(cpl, rank, idx, st, abnd, cplarridx, err)
    type(fd4_couple),     intent(inout)        :: cpl          !! couple context
    integer(i_k),         intent(in)           :: rank         !! owner of this coupling array
    integer(i_k),         intent(in)           :: idx          !! field index in vartab
    integer(i_k),         intent(in)           :: st           !! time step index
    integer(i_k),         intent(in)           :: abnd(3,2)    !! bounds of the coupling array in global coordinates
    integer(i_k),         intent(inout)        :: cplarridx    !! local couple array index (>0), unchanged if not local
    integer(i_k),         intent(out)          :: err          !! error status: 0...ok
    
    integer(i_k) :: i3(3), i, abnd_idx

    err = -1
    FD4_L1_ASSERT(cpl%state<CPL_CRT,'cpl not created/initialized')
    FD4_L1_ASSERT(cpl%state>CPL_ADD,'cpl has already local data fields')
    FD4_L1_ASSERT(cpl%ins_rank>rank,'need increasing or equal ranks')
    
    if(rank > cpl%ins_rank) then
      
      !! did not see this rank before !!
      
      i3(1) = idx
      i3(2) = st
      call stack3_push(cpl%desc_abnd, abnd(1:3,1))
      i3(3) = cpl%desc_abnd%pos ! index in desc_abnd
      call stack3_push(cpl%desc_abnd, abnd(1:3,2))
      call stack3_push(cpl%desc_var, i3)
      cpl%desc_rank(rank) = cpl%desc_var%pos
      cpl%ins_rank = rank
      
    else  ! rank == cpl%ins_rank
      
      !! rank already present !!
      
      ! check if abnd is already in desc_abnd: loop over all abnd of this rank
      abnd_idx = -1
      do i=cpl%desc_var%a(3,cpl%desc_rank(rank)), cpl%desc_abnd%pos, 2
        if( cpl%desc_abnd%a(1,i  ) /= abnd(1,1) ) cycle
        if( cpl%desc_abnd%a(2,i  ) /= abnd(2,1) ) cycle
        if( cpl%desc_abnd%a(3,i  ) /= abnd(3,1) ) cycle
        if( cpl%desc_abnd%a(1,i+1) /= abnd(1,2) ) cycle
        if( cpl%desc_abnd%a(2,i+1) /= abnd(2,2) ) cycle
        if( cpl%desc_abnd%a(3,i+1) /= abnd(3,2) ) cycle
        abnd_idx = i  ! hit!
        exit
      end do
      if(abnd_idx<=0) then
        ! found no matching abnd: add new entry
        call stack3_push(cpl%desc_abnd, abnd(1:3,1))
        i3(3) = cpl%desc_abnd%pos ! index in desc_abnd
        call stack3_push(cpl%desc_abnd, abnd(1:3,2))
      else
        ! found matching abnd: referencing this abnd entry
        i3(3) = abnd_idx
      end if
      i3(1) = idx
      i3(2) = st
      call stack3_push(cpl%desc_var, i3)
      
    end if

    cpl%desc_rank(rank+1) = cpl%desc_var%pos+1
    
    if(rank == cpl%domain%mpi%rank) then
      ! this is a local array
      cpl%ncplarrays = cpl%ncplarrays + 1
      cplarridx = cpl%ncplarrays
    end if

    cpl%state = CPL_ADD
    err = 0
  
  end subroutine fd4_couple_add_array


  !! Add a local 3D coupling array to the FD4 couple context.
  !!
  !! Only allowed after $fd4_couple_add_array and before $fd4_couple_commit.
  !! Needs to be called for all local coupling arrays which have been added via
  !! $fd4_couple_add_array with $rank parameter == local rank.
  !!
  subroutine fd4_couple_set_local_3D_array(cpl, cplarridx, array, offset, err)
    type(fd4_couple),     intent(inout)  :: cpl          !! couple context
    integer(i_k),         intent(in)     :: cplarridx    !! local couple array index, value returned by $fd4_couple_add_array
    real(r_k), pointer                   :: array(:,:,:) !! data array
    integer(i_k),         intent(in)     :: offset(3)    !! spatial offset of the data in array, e.g. (/1,1,1/)
    integer(i_k),         intent(out)    :: err          !! error status: 0...ok
  
    err = -1
    FD4_L1_ASSERT(cpl%state<CPL_ADD,'add variables first')
    FD4_L1_ASSERT(cpl%state>CPL_LADD,'couple already commited')
    FD4_L3_ASSERT(cplarridx<1,'cplarridx <1')
    FD4_L3_ASSERT(cplarridx>cpl%ncplarrays,'cplarridx too large')
    
    if(cpl%state == CPL_ADD) then
      allocate(cpl%desc_local(cpl%ncplarrays))
      cpl%state = CPL_LADD
    end if
    
    cpl%desc_local(cplarridx)%offset = offset
    cpl%desc_local(cplarridx)%array => array
    cpl%desc_local(cplarridx)%array4 => NULL()
    cpl%desc_local(cplarridx)%offset4 = 1

    err = 0

  end subroutine fd4_couple_set_local_3D_array


  !! Add a local 4D coupling array to the FD4 couple context.
  !!
  !! Only allowed after $fd4_couple_add_array and before $fd4_couple_commit.
  !! Needs to be called for all local coupling arrays which have been added via
  !! $fd4_couple_add_array with $rank parameter == local rank.
  !!
  subroutine fd4_couple_set_local_4D_array(cpl, cplarridx, array4, offset, offset4, err)
    type(fd4_couple),     intent(inout)  :: cpl             !! couple context
    integer(i_k),         intent(in)     :: cplarridx       !! local couple array index, value returned by $fd4_couple_add_array
    real(r_k), pointer                   :: array4(:,:,:,:) !! data array
    integer(i_k),         intent(in)     :: offset(3)       !! spatial offset of the data in array, e.g. (/1,1,1/)
    integer(i_k),         intent(in)     :: offset4         !! offset of the data in the 4th dimension
    integer(i_k),         intent(out)    :: err             !! error status: 0...ok
  
    err = -1
    FD4_L1_ASSERT(cpl%state<CPL_ADD,'add variables first')
    FD4_L1_ASSERT(cpl%state>CPL_LADD,'couple already commited')
    FD4_L3_ASSERT(cplarridx<1,'cplarridx <1')
    FD4_L3_ASSERT(cplarridx>cpl%ncplarrays,'cplarridx too large')
    
    if(cpl%state == CPL_ADD) then
      allocate(cpl%desc_local(cpl%ncplarrays))
      cpl%state = CPL_LADD
    end if
    
    cpl%desc_local(cplarridx)%offset = offset
    cpl%desc_local(cplarridx)%array => NULL()
    cpl%desc_local(cplarridx)%array4 => array4
    cpl%desc_local(cplarridx)%offset4 = offset4

    err = 0

  end subroutine fd4_couple_set_local_4D_array
  

  !! Change a local 3D coupling array to the FD4 couple context.
  !!
  !! Only allowed after $fd4_couple_commit.
  !!
  subroutine fd4_couple_chg_local_3D_array(cpl, cplarridx, array, err)
    type(fd4_couple),     intent(inout)  :: cpl             !! couple context
    integer(i_k),         intent(in)     :: cplarridx       !! local couple array index, value returned by $fd4_couple_add_array
    real(r_k), pointer                   :: array(:,:,:)    !! data array
    integer(i_k),         intent(out)    :: err             !! error status: 0...ok
  
    integer(i_k) :: idesc, nbins
    
    err = -1
    FD4_L1_ASSERT(cpl%state/=CPL_CMMT,'commit cpl first')
    FD4_L3_ASSERT(cplarridx<1,'cplarridx <1')
    FD4_L3_ASSERT(cplarridx>cpl%ncplarrays,'cplarridx too large')
    
    idesc = cplarridx + cpl%desc_rank(cpl%domain%mpi%rank) - 1 ! index of this couple array in cpl%desc_var
    nbins = cpl%domain%vartab( cpl%desc_var%a(1,idesc) )%nbins
    FD4_L1_ASSERT(nbins/=1,'cannot set 3D couple array for 4D variable')
    
    cpl%desc_local(cplarridx)%array => array
    cpl%desc_local(cplarridx)%array4 => NULL()

    err = 0

  end subroutine fd4_couple_chg_local_3D_array


  !! Change a local 4D coupling array to the FD4 couple context.
  !!
  !! Only allowed after $fd4_couple_commit.
  !!
  subroutine fd4_couple_chg_local_4D_array(cpl, cplarridx, array4, err)
    type(fd4_couple),     intent(inout)  :: cpl             !! couple context
    integer(i_k),         intent(in)     :: cplarridx       !! local couple array index, value returned by $fd4_couple_add_array
    real(r_k), pointer                   :: array4(:,:,:,:) !! data array
    integer(i_k),         intent(out)    :: err             !! error status: 0...ok
  
    integer(i_k) :: idesc, nbins

    err = -1
    FD4_L1_ASSERT(cpl%state/=CPL_CMMT,'commit cpl first')
    FD4_L3_ASSERT(cplarridx<1,'cplarridx <1')
    FD4_L3_ASSERT(cplarridx>cpl%ncplarrays,'cplarridx too large')

    idesc = cplarridx + cpl%desc_rank(cpl%domain%mpi%rank) - 1 ! index of this couple array in cpl%desc_var
    nbins = cpl%domain%vartab( cpl%desc_var%a(1,idesc) )%nbins
    FD4_L1_ASSERT(nbins==1,'cannot set 4D couple array for 3D variable')
    
    cpl%desc_local(cplarridx)%array => NULL()
    cpl%desc_local(cplarridx)%array4 => array4

    err = 0

  end subroutine fd4_couple_chg_local_4D_array


  !! Commit FD4 couple context.
  !!
  subroutine fd4_couple_commit(cpl, err, opt_face_overlap)
    type(fd4_couple),  intent(inout), target  :: cpl          !! couple context
    integer(i_k),      intent(out)            :: err          !! error status: 0...ok
    !! Do the couple arrays for face variables overlap at the inner boundaries (one cell in face direction)? Default is yes.
    logical, intent(in), optional             :: opt_face_overlap

    
    integer(i_k) :: nbins
    type(fd4_mpi), pointer :: mpi
    integer(i_k) :: idx, idesc, ilocal, dest
    integer(i_k) :: realabnd(0:3,2)
    integer(i_k) :: bbnd(3,2),tbnd(3,2),face,intersec(3,2)
    integer(i_k) :: i,j,k
    integer(i_k) :: myabnd(3,2), tabnd(3,2)
    integer :: blocktype
    integer(i_k) :: dim, facetype
    logical :: facevar_present(0:3)
    type(fd4_domain), pointer :: domain
    integer(i_k), parameter :: CPL_NOFACE=1, CPL_FACE=2, CPL_NO=3

    err = -1
    FD4_L1_ASSERT(cpl%state<CPL_ADD,'add couple vars first')
    FD4_L1_ASSERT(cpl%state>=CPL_CMMT,'cpl already commited')
    
    domain => cpl%domain
    mpi => domain%mpi
    
    if(present(opt_face_overlap)) then
      cpl%face_overlap = opt_face_overlap
    else
      cpl%face_overlap = .true.
    end if

    intersec(:,1) = 0
    intersec(:,2) = -1
    facevar_present = .false.
    cpl%faceoffset(1:3) = 0
    
#ifdef FD4_COUPLE_DEBUG
    write(300+mpi%rank,'(A,I5)')  '---- fd4_couple_commit ',cpl%desc_var%pos
#endif

    ! check params
    do ilocal=1,cpl%ncplarrays  ! loop over local couple arrays
      idesc = ilocal + cpl%desc_rank(mpi%rank) - 1 ! index of this couple array in cpl%desc_var
      idx = cpl%desc_var%a(1,idesc)
      nbins = domain%vartab(idx)%nbins
      if( nbins == 1 .and. .not. associated(cpl%desc_local(ilocal)%array)) then
        FD4_L1_ERROR('array not associated at '//i2s(ilocal))
      end if
      if( nbins >  1 .and. .not. associated(cpl%desc_local(ilocal)%array4)) then
        FD4_L1_ERROR('array4 not associated at '//i2s(ilocal))
      end if
      if(associated(cpl%desc_local(ilocal)%array) .and. associated(cpl%desc_local(ilocal)%array4)) then
        FD4_L1_ERROR('array and array4 associated at '//i2s(ilocal))
      end if
      ! get bounds of the couple array within the spatial grid
      myabnd(1:3,1) = cpl%desc_abnd%a(1:3,cpl%desc_var%a(3,idesc)  )
      myabnd(1:3,2) = cpl%desc_abnd%a(1:3,cpl%desc_var%a(3,idesc)+1)
      if( ANY(myabnd(1:3,2) - myabnd(1:3,1) < 0 ) ) then
        FD4_L1_ERROR('abnd size is <1 for '//i2s(ilocal))
      end if
!#if ( FD4_VERBOSE_LEVEL >= 3 )
      ! get real/allocated bounds of the couple array
      if( nbins > 1 ) then
        realabnd(0:3,1) = lbound(cpl%desc_local(ilocal)%array4)
        realabnd(0:3,2) = ubound(cpl%desc_local(ilocal)%array4)
      else
        realabnd(0,1:2) = 0
        realabnd(1:3,1) = lbound(cpl%desc_local(ilocal)%array)
        realabnd(1:3,2) = ubound(cpl%desc_local(ilocal)%array)
      end if
      ! check array size for 4th dimension
      if( nbins > 1 ) then
        if( realabnd(0,1) > cpl%desc_local(ilocal)%offset4 ) then
          FD4_L1_ERROR('desc_local%offset4 too small at '//i2s(ilocal))
        end if
        if( realabnd(0,2) < cpl%desc_local(ilocal)%offset4 + nbins - 1 ) then
          FD4_L1_ERROR('desc_local%array4 too small at '//i2s(ilocal))
        end if
      end if      
      ! check array size for spatial dimensions      
      if( ANY( realabnd(1:3,1) > cpl%desc_local(ilocal)%offset ) ) then
        FD4_L1_ERROR('desc%offset too small at '//i2s(idesc))
      end if
      if( ANY( realabnd(1:3,2) < cpl%desc_local(ilocal)%offset + myabnd(1:3,2) - myabnd(1:3,1) ) ) then
        FD4_L1_ERROR('desc%array too small at '//i2s(idesc))
      end if
!#endif
      ! get bounds
      if( intersec(1,1) > intersec(1,2) ) then
        ! 1st time here
        intersec = myabnd
      else
        ! add this coupling array to the union of all coupling arrays
        ! (yes, variable is called intersection here...)
        intersec(1:3,1)=MIN(intersec(1:3,1),myabnd(1:3,1))
        intersec(1:3,2)=MAX(intersec(1:3,2),myabnd(1:3,2))
      end if
    end do
    
    ! check for face variables in all couple arrays
    do idesc=1,cpl%desc_var%pos  ! loop over all couple arrays (global)
      idx = cpl%desc_var%a(1,idesc)
      facevar_present(domain%vartab(idx)%facevar) = .true.
      if(domain%vartab(idx)%facevar>0 .and. domain%vartab(idx)%facevar<4) then
        cpl%faceoffset(domain%vartab(idx)%facevar) = 1
      end if
    end do  

    cpl%loopbnd(1:3,1) =  0
    cpl%loopbnd(1:3,2) = -1

    if(cpl%ncplarrays>0) then

      ! reduce the number of blocks to loop by taking the
      ! union of all coupling arrays into account
      ! 1. intersect this union with the domain size
      intersec(1:3,1)=MAX(intersec(1:3,1),domain%dbnd(1:3,1))
      intersec(1:3,2)=MIN(intersec(1:3,2),domain%dbnd(1:3,2))
      ! 2. now calculate the loop bounds
      do dim=1,3
        ! try to guess the starting block
        i = ( domain%blnum(dim) * (intersec(dim,1)-domain%dbnd(dim,1)+1) ) / (domain%dbnd(dim,2)-domain%dbnd(dim,1)+1)
        i = MIN(MAX(1,i),domain%blnum(dim))
        ! find real start
        do
          if(domain%d(dim)%bl(i)>intersec(dim,1)) then
            i=i-1
          else if(domain%d(dim)%bl(i+1)<=intersec(dim,1)) then
            i=i+1
          else
            exit
          end if
        end do
        ! shift the start forward if
        ! * facevar present 
        ! * block start and intersection start exactly at the same position 
        if(facevar_present(dim) .and. intersec(dim,1)==domain%d(dim)%bl(i) .and. i>1) i=i-1
        cpl%loopbnd(dim,1) = i 
        ! try to guess the end block
        i = ( domain%blnum(dim) * (intersec(dim,2)-domain%dbnd(dim,1)+1) ) / (domain%dbnd(dim,2)-domain%dbnd(dim,1)+1)
        i = MIN(MAX(1,i),domain%blnum(dim))
        ! find real end
        do
          if(domain%d(dim)%bl(i)>intersec(dim,2)) then
            i=i-1
          else if(domain%d(dim)%bl(i+1)<=intersec(dim,2)) then
            i=i+1
          else
            exit
          end if
        end do    
        cpl%loopbnd(dim,2) = i
      end do


      allocate(cpl%arraytypes(cpl%ncplarrays, cpl%loopbnd(1,1):cpl%loopbnd(1,2), &
                                              cpl%loopbnd(2,1):cpl%loopbnd(2,2), &
                                              cpl%loopbnd(3,1):cpl%loopbnd(3,2),   4  ) )


      cpl%arraytypes = MPI_DATATYPE_NULL
      cpl%arraytypes(:,:,:,:,CPL_GET_CASE) = CPL_GET_CASE_SKIP

      !! pre-build types for the coupling array (subsections which correspond to an FD4 block each)

      ! loop over all FD4 blocks which intersect with local couple arrays
      do k=cpl%loopbnd(3,1),cpl%loopbnd(3,2)
        do j=cpl%loopbnd(2,1),cpl%loopbnd(2,2)
          do i=cpl%loopbnd(1,1),cpl%loopbnd(1,2)

            ! loop over local couple arrays        
            do ilocal=1,cpl%ncplarrays

              ! set some convenience vars
              idesc = ilocal + cpl%desc_rank(mpi%rank) - 1 ! index of this couple array in cpl%desc_var
              ! get bounds of the couple array within the spatial grid
              myabnd(1:3,1) = cpl%desc_abnd%a(1:3,cpl%desc_var%a(3,idesc)  )
              myabnd(1:3,2) = cpl%desc_abnd%a(1:3,cpl%desc_var%a(3,idesc)+1)
              idx = cpl%desc_var%a(1,idesc)
              face = domain%vartab(idx)%facevar
              nbins = domain%vartab(idx)%nbins

              ! bounds of this non-existing block
              bbnd(1,:) = (/ domain%d(1)%bl(i) , domain%d(1)%bl(i+1) - 1 /)
              bbnd(2,:) = (/ domain%d(2)%bl(j) , domain%d(2)%bl(j+1) - 1 /)
              bbnd(3,:) = (/ domain%d(3)%bl(k) , domain%d(3)%bl(k+1) - 1 /)


              if(face==0) then
              !!! build datatype without face

                tbnd = bbnd
                tabnd = myabnd
                call fd4_couple_build_cplarray_type(cpl%desc_local(ilocal), tabnd, tbnd, nbins, blocktype, err )
                FD4_L3_ASSERT(err/=0,'fd4_couple_build_cplarray_type failed')
                if(blocktype /= MPI_DATATYPE_NULL) then
                  ! save this type in the arraytypes array
                  cpl%arraytypes(ilocal, i, j, k, CPL_GET) = blocktype
                  cpl%arraytypes(ilocal, i, j, k, CPL_PUT) = blocktype
                  cpl%arraytypes(ilocal, i, j, k, CPL_GET_CASE) = CPL_GET_CASE_NORMAL
#ifdef FD4_COUPLE_DEBUG
                  write(300+mpi%rank,'(A,I5,I5,I5,A,I5,I5,I5,A,I8,A)') &
                      'CREA ',idx,cpl%desc_var%a(2,idesc),ilocal,' p:',i,j,k,' t:',blocktype,'  no face'
#endif              
                end if
              end if
              
              if(face>0 .and. face<4 .and. IAND(cpl%cpldir,FD4_CPL_PUT)/=0) then
              !!! build datatype with face for fd4_couple_put

                ! additional cell plane if face var
                tbnd = bbnd
                tbnd(face,2) = tbnd(face,2) + 1
                tabnd = myabnd

                if(cpl%face_overlap) then
                  if(myabnd(face,1)==tbnd(face,2)) then
                    ! CASE B: first cell of couple array == last cell of block:
                    ! do not transfer anything, left neighbor couple array will transfer all
                    facetype = CPL_NO ! 3
                  else if(myabnd(face,2)<tbnd(face,2)) then
                    ! CASE A: last cell of couple array is left of last cell of block:
                    ! do not transfer last cell _of_the_couple_array_, will be transfered by right neighbor couple array
                    facetype = CPL_NOFACE ! 1
                    tabnd(face,2) = tabnd(face,2) - 1
                  else
                    ! CASE C: last cell of couple array >= last cell of block:
                    ! transfer all incl. last cell, because right neighbor won't do that
                    facetype = CPL_FACE ! 2
                  end if
                else
                  facetype = CPL_FACE
                end if

                if(facetype /= CPL_NO) then
                  call fd4_couple_build_cplarray_type(cpl%desc_local(ilocal), tabnd, tbnd, nbins, blocktype, err )
                  FD4_L3_ASSERT(err/=0,'fd4_couple_build_cplarray_type failed')
                  if(blocktype /= MPI_DATATYPE_NULL) then
                    ! save this type in the arraytypes array
                    cpl%arraytypes(ilocal, i, j, k, CPL_PUT) = blocktype
#ifdef FD4_COUPLE_DEBUG
                    write(300+mpi%rank,'(A,I5,I5,I5,A,I5,I5,I5,A,I8,A,I2)') &
                        'CREA ',idx,cpl%desc_var%a(2,idesc),ilocal,' p:',i,j,k,' t:',blocktype,'  face put', facetype
#endif              
                  end if
                end if
              end if
              
              if(face>0 .and. face<4 .and. IAND(cpl%cpldir,FD4_CPL_GET)/=0) then
              !!! build datatype with face for fd4_couple_get

                ! additional cell plane if face var
                tbnd = bbnd
                tbnd(face,2) = tbnd(face,2) + 1
                tabnd = myabnd

                ! build type with additional face cell
                call fd4_couple_build_cplarray_type(cpl%desc_local(ilocal), tabnd, tbnd, nbins, blocktype, err )
                FD4_L3_ASSERT(err/=0,'fd4_couple_build_cplarray_type failed')
                if(blocktype /= MPI_DATATYPE_NULL) then
                  ! save this type in the arraytypes array
                  cpl%arraytypes(ilocal, i, j, k, CPL_GET) = blocktype
#ifdef FD4_COUPLE_DEBUG
                  write(300+mpi%rank,'(A,I5,I5,I5,A,I5,I5,I5,A,I8,A)') &
                      'CREA ',idx,cpl%desc_var%a(2,idesc),ilocal,' p:',i,j,k,' t:',blocktype,'  face get CPL_FACE'
#endif              

                  !! switch between cases depending on position of couple array / FD4 block
                  if( myabnd(face,2) == tbnd(face,1) ) then
                    ! case B: last cell of couple array is first cell of FD4 block,
                    ! left neighbor could provide the data if present
                    ! (fd4_couple_get decides at run-time whether to transfer data or not)
                    cpl%arraytypes(ilocal, i, j, k, CPL_GET_CASE) = CPL_GET_CASE_CHECK_L
                  elseif( myabnd(face,2) > tbnd(face,2) ) then
                    ! case C: last cell of couple array is right outside the FD4 block,
                    ! right neighbor could provide additional face cell if present
                    cpl%arraytypes(ilocal, i, j, k, CPL_GET_CASE) = CPL_GET_CASE_CHECK_R
                    ! build type without additional face cell _of_the_FD4_block_
                    ! (fd4_couple_get decides at run-time which MPI type to take)
                    tbnd = bbnd
                    call fd4_couple_build_cplarray_type(cpl%desc_local(ilocal), tabnd, tbnd, nbins, blocktype, err )
                    FD4_L3_ASSERT(err/=0,'fd4_couple_build_cplarray_type failed')
                    if(blocktype /= MPI_DATATYPE_NULL) then
                      ! save this type in the arraytypes array
                      cpl%arraytypes(ilocal, i, j, k, CPL_GET_NOFACE) = blocktype
#ifdef FD4_COUPLE_DEBUG
                      write(300+mpi%rank,'(A,I5,I5,I5,A,I5,I5,I5,A,I8,A)') &
                          'CREA ',idx,cpl%desc_var%a(2,idesc),ilocal,' p:',i,j,k,' t:',blocktype,'  face get CPL_NOFACE'
#endif              
                    end if
                  else
                    ! case A: last cell of couple array resides in FD4 block but is not first cell
                    ! (no dependence on neighbor blocks)
                    cpl%arraytypes(ilocal, i, j, k, CPL_GET_CASE) = CPL_GET_CASE_NORMAL
                  end if

                end if
              end if
              !!! face

            end do ! local couple arrays

          end do
        end do
      end do
    
    end if

    
    ! for each rank: calculate union of the bounds of all couple arrays
    allocate(cpl%rankbnd(3,2,0:mpi%nprocs-1))
    ! initialize with empty (negative) bounds
    do dest=0,mpi%nprocs-1
      cpl%rankbnd(1:3,1,dest) =  0
      cpl%rankbnd(1:3,2,dest) = -1
      do idesc=cpl%desc_rank(dest), cpl%desc_rank(dest+1) - 1
        myabnd(1:3,1) = cpl%desc_abnd%a(1:3,cpl%desc_var%a(3,idesc)  )
        myabnd(1:3,2) = cpl%desc_abnd%a(1:3,cpl%desc_var%a(3,idesc)+1)
        if( cpl%rankbnd(1,1,dest) > cpl%rankbnd(1,2,dest) ) then
          ! first block of rank dest
          cpl%rankbnd(1:3,1:2,dest) = myabnd(1:3,1:2)
        else
          cpl%rankbnd(1:3,1,dest)=MIN(cpl%rankbnd(1:3,1,dest),myabnd(1:3,1))
          cpl%rankbnd(1:3,2,dest)=MAX(cpl%rankbnd(1:3,2,dest),myabnd(1:3,2))
        end if      
      end do
    end do
    
    cpl%state = CPL_CMMT
    err = 0

  end subroutine fd4_couple_commit
  


  !! Build an MPI data type for a couple array and an FD4 block. Internal routine.
  subroutine fd4_couple_build_cplarray_type(desc_local, abnd, bbnd, nbins, datatype, err )
    type(fd4_couple_local_desc), intent(in)  :: desc_local !! description of couple local array
    integer(i_k),          intent(in)        :: abnd(3,2)  !! bounds of couple array
    integer(i_k),          intent(in)        :: bbnd(3,2)  !! bounds of this (locally non-existing) block
    integer(i_k),          intent(in)        :: nbins      !! number of bins of the variable in desc
    integer,               intent(out)       :: datatype   !! the MPI data type
    integer(i_k),          intent(out)       :: err        !! error status: 0...ok
    
    integer(i_k) :: sbnd(3,2),intersec(3,2),stride(1,2),ext(0:3)

    err = -1
    datatype = MPI_DATATYPE_NULL

    ! calculate intersection of block with array
    intersec(1:3,1)=MAX(bbnd(1:3,1),abnd(1:3,1))
    intersec(1:3,2)=MIN(bbnd(1:3,2),abnd(1:3,2))

    ! check if intersection is not null
    if(intersec(1,1)<=intersec(1,2) .and. &
       intersec(2,1)<=intersec(2,2) .and. &
       intersec(3,1)<=intersec(3,2) ) then

      if(nbins==1) then
        stride(1,1:2) = 1
        ext(0) = 1
        ext(1:3) = ubound(desc_local%array) - lbound(desc_local%array) + 1
      else
        stride(1,1) = desc_local%offset4
        stride(1,2) = desc_local%offset4 + nbins - 1
        stride(1,:) = stride(1,:) - lbound(desc_local%array4,1) + 1
        ext(0:3) = ubound(desc_local%array4) - lbound(desc_local%array4) + 1
      end if

      ! bounds of subset in array which respects to the current block, array content starts at given offset
      sbnd(1:3,1) = intersec(1:3,1) - abnd(1:3,1) + desc_local%offset(1:3)
      sbnd(1:3,2) = intersec(1:3,2) - abnd(1:3,1) + desc_local%offset(1:3)
      ! go, build the datatype for array section sbnd
      call fd4_comm_build_array_type(ext, stride, sbnd, datatype, err)
      FD4_L3_ASSERT(err/=0,'fd4_comm_build_array_type failed')

    end if      

    err = 0
    
  end subroutine fd4_couple_build_cplarray_type
  


  
  !! Delete the couple context and all its memory.
  !!
  subroutine fd4_couple_delete(cpl, err)
    type(fd4_couple),  intent(inout)   :: cpl          !! couple context
    integer(i_k),      intent(out)     :: err          !! error status: 0...ok

    integer(i_k) :: i, j ,k, icplarrays
    integer, pointer :: cpl_get_type, cpl_put_type
    
    err = -1

    FD4_L1_ASSERT(.not. associated(cpl%domain),'cpl invalid')

    ! deallocate MPI types of communication partners
    call rbtree_fd4_neigh_clear(cpl%neighstore,fd4_neigh_deallocate_ptr)
    
    ! free basic types for the couple array
    if(associated(cpl%arraytypes)) then
      do k=cpl%loopbnd(3,1),cpl%loopbnd(3,2)
        do j=cpl%loopbnd(2,1),cpl%loopbnd(2,2)
          do i=cpl%loopbnd(1,1),cpl%loopbnd(1,2)
            do icplarrays=1,cpl%ncplarrays
              cpl_put_type => cpl%arraytypes(icplarrays, i, j, k, CPL_PUT)
              cpl_get_type => cpl%arraytypes(icplarrays, i, j, k, CPL_GET)
              if( cpl_put_type /= MPI_DATATYPE_NULL ) then
                if( cpl_put_type == cpl_get_type) then
                  ! cpl_put data type equals cpl_get data type (see fd4_couple_commit)
                  cpl_get_type = MPI_DATATYPE_NULL
                end if
                call MPI_Type_free( cpl_put_type, err)
                err = err - MPI_SUCCESS
                FD4_L3_ASSERT(err/=0,'MPI_Type_free failed')
              end if
              if( cpl_get_type /= MPI_DATATYPE_NULL ) then
                call MPI_Type_free( cpl_get_type, err)
                err = err - MPI_SUCCESS
                FD4_L3_ASSERT(err/=0,'MPI_Type_free failed')
              end if
              cpl_get_type => cpl%arraytypes(icplarrays, i, j, k, CPL_GET_NOFACE)
              if( cpl_get_type /= MPI_DATATYPE_NULL ) then
                call MPI_Type_free( cpl_get_type, err)
                err = err - MPI_SUCCESS
                FD4_L3_ASSERT(err/=0,'MPI_Type_free failed')
              end if
            end do
          end do          
        end do
      end do
      deallocate(cpl%arraytypes)
    end if
    
    if(associated(cpl%rankbnd))    deallocate(cpl%rankbnd)
    if(associated(cpl%desc_rank))  deallocate(cpl%desc_rank)
    if(associated(cpl%desc_local)) deallocate(cpl%desc_local)
    call stack3_delete(cpl%desc_var)
    call stack3_delete(cpl%desc_abnd)

    NULLIFY(cpl%domain)
    cpl%state = CPL_NEW    
    err = 0
  
  end subroutine fd4_couple_delete
  
    

  !! Put data to the framework. Each process contributes several blocks of
  !! data. $desc%abnd has to be set in all processes correctly. 
  !! You must supply the same $desc as in $fd4_couple_create, the only
  !! allowed change is the pointer to the data ($desc%array).
  !!
  !! *TODO*:
  !!  - see limitations above
  !!  - much more parameter checking
  !!  - check if we receive a cell twice (from different processes) ??
  !!
  subroutine fd4_couple_put(cpl, err)
    type(fd4_couple),     intent(inout),target :: cpl          !! couple context
    integer(i_k),         intent(out)          :: err          !! error status: 0...ok
    
    integer(i_k) :: ilocal, nbins
    type(fd4_mpi), pointer :: mpi
    integer(i8k) :: index
    integer(i_k) :: idx, st, idesc, lb(4)
    integer(i_k) :: i,j,k,bbnd(3,2),sbnd(3,2),intersec(3,2),face,stride(1,2),fext(3),ext(3),bnd4th(2)
    real(r_k), pointer :: minptr
    integer(i_k) :: myabnd(3,2)
    integer :: dest, blocktype, request
    integer(MPI_ADDRESS_KIND) :: disp
    type(fd4_block), pointer :: b
    type(fd4_iter), target  :: iter
    logical :: is_min
    type(fd4_domain), pointer :: domain
    real(r_k), pointer :: mindisparray             ! pointer to couple array with min. address
    integer(MPI_ADDRESS_KIND) :: mindisp           ! address of couple array with min. address
    type(fd4_mpitype), pointer :: desttype
#if ( FD4_VERBOSE_LEVEL >= 3 )
    integer(i_k) :: realabnd(0:3,2)
#endif
#ifdef FD4_COUPLE_DEBUG
    character(1) :: case_str
#endif    
#ifdef FD4_COUPLE_CHECK_RECV_SIZE
    integer :: status(MPI_STATUS_SIZE)
    integer :: msgsize_exp, msgsize_recv
#endif
    type(fd4_neigh), pointer :: n, deleted
    type(rbtree_fd4_neigh_iterator) :: niter

    err = -1
    FD4_L1_ASSERT(cpl%state/=CPL_CMMT,'cpl not commited')
    if(IAND(cpl%cpldir,FD4_CPL_PUT)==0) then
      FD4_L1_ERROR('fd4_couple_put not possible')
    end if

    domain => cpl%domain
    mpi => domain%mpi
    minptr => NULL()
    mindisp = HUGE(1_MPI_ADDRESS_KIND)
    nullify(mindisparray)
    
    ! loop over local couple arrays:
    ! check params and get array with minimum displacement
    do ilocal=1,cpl%ncplarrays
      idesc = ilocal + cpl%desc_rank(mpi%rank) - 1 ! index of this couple array in cpl%desc_var
      idx = cpl%desc_var%a(1,idesc)
      nbins = domain%vartab(idx)%nbins
      if( nbins == 1 .and. .not. associated(cpl%desc_local(ilocal)%array)) then
        FD4_L1_ERROR('desc_local%array not associated at '//i2s(ilocal))
      end if
      if( nbins >  1 .and. .not. associated(cpl%desc_local(ilocal)%array4)) then
        FD4_L1_ERROR('desc_local%array4 not associated at '//i2s(ilocal))
      end if
      ! find min displacement
      if( nbins > 1 ) then
        lb(1:4) = lbound(cpl%desc_local(ilocal)%array4)
        call MPI_Get_address( cpl%desc_local(ilocal)%array4(lb(1),lb(2),lb(3),lb(4)), disp, err)
        if(disp < mindisp) then
          mindisp = disp
          mindisparray => cpl%desc_local(ilocal)%array4(lb(1),lb(2),lb(3),lb(4))
        end if
      else
        lb(1:3) = lbound(cpl%desc_local(ilocal)%array)
        call MPI_Get_address( cpl%desc_local(ilocal)%array(lb(1),lb(2),lb(3)), disp, err)
        if(disp < mindisp) then
          mindisp = disp
          mindisparray => cpl%desc_local(ilocal)%array(lb(1),lb(2),lb(3))
        end if
      end if
      ! save array address
      cpl%desc_local(ilocal)%disp = disp
#if ( FD4_VERBOSE_LEVEL >= 3 )
      ! get real/allocated bounds of the couple array
      if( nbins > 1 ) then
        realabnd(0:3,1) = lbound(cpl%desc_local(ilocal)%array4)
        realabnd(0:3,2) = ubound(cpl%desc_local(ilocal)%array4)
      else
        realabnd(0,1:2) = 0
        realabnd(1:3,1) = lbound(cpl%desc_local(ilocal)%array)
        realabnd(1:3,2) = ubound(cpl%desc_local(ilocal)%array)
      end if
      ! check array size for 4th dimension
      if( nbins > 1 ) then
        if( realabnd(0,1) > cpl%desc_local(ilocal)%offset4 ) then
          FD4_L1_ERROR('desc_local%offset4 too small at '//i2s(ilocal))
        end if
        if( realabnd(0,2) < cpl%desc_local(ilocal)%offset4 + nbins - 1 ) then
          FD4_L1_ERROR('desc_local%array4 too small at '//i2s(ilocal))
        end if
      end if      
      ! get bounds of the couple array within the spatial grid
      myabnd(1:3,1) = cpl%desc_abnd%a(1:3,cpl%desc_var%a(3,idesc)  )
      myabnd(1:3,2) = cpl%desc_abnd%a(1:3,cpl%desc_var%a(3,idesc)+1)
      ! check array size for spatial dimensions      
      if( ANY( realabnd(1:3,1) > cpl%desc_local(ilocal)%offset ) ) then
        FD4_L1_ERROR('desc%offset too small at '//i2s(idesc))
      end if
      if( ANY( realabnd(1:3,2) < cpl%desc_local(ilocal)%offset + myabnd(1:3,2) - myabnd(1:3,1) ) ) then
        FD4_L1_ERROR('desc%array too small at '//i2s(idesc))
      end if
#endif
    end do


#ifdef FD4_COUPLE_DEBUG
    write(300+mpi%rank,'(A,I5)')  '---- fd4_couple_put ',cpl%ncplarrays
#endif


    mpi%requests%pos = 0
  
    
    !
    !  1. send data
    !

    ! loop over all blocks of the domain
    ! to build an MPI datatype to send subsections
    ! from myarray to the blocks of other processes
    do k=cpl%loopbnd(3,1),cpl%loopbnd(3,2)
      do j=cpl%loopbnd(2,1),cpl%loopbnd(2,2)
        do i=cpl%loopbnd(1,1),cpl%loopbnd(1,2)
        
          ! cycle the loop if block does not exist
          dest = domain%blkmap(i,j,k,domain%blkmap_now)
          if(dest == -1 ) cycle

          ! cycle if this is my block
          if(dest == mpi%rank) cycle
          
          ! loop over local couple arrays
          do ilocal=1,cpl%ncplarrays
            idesc = ilocal + cpl%desc_rank(mpi%rank) - 1 ! index of this couple array in cpl%desc_var
            
            ! get the matching type from the array of pre-built datatypes
            blocktype = cpl%arraytypes(ilocal, i, j, k, CPL_PUT)

            ! check if type exists
            if(blocktype /= MPI_DATATYPE_NULL) then

              index = INT( fd4_domain_ijk2if(i,j,k), i8k)   ! index for sorting
              
              ! check if a neigh type for this rank already exists, if not, create
              n => rbtree_fd4_neigh_get(cpl%neighstore, dest)
              if(.not. associated(n)) then
                allocate(n, stat=err)
                FD4_L3_ASSERT(err/=0,'allocate neigh')
                call rbtree_fd4_neigh_insert (cpl%neighstore, dest, n)
              end if

              call fd4_mpitype_add(n%mpitype(FD4_NEIGH_CPL), index, 1, cpl%desc_local(ilocal)%disp, blocktype)
#ifdef FD4_COUPLE_DEBUG
              call MPI_Type_size(blocktype, request, err)
              write(300+mpi%rank,'(A,I5,I5,I5,A,I5,I5,I5,A,I10,I7,I12,I7,I26)')   &
                  'SEND ',cpl%desc_var%a(1,idesc),cpl%desc_var%a(2,idesc),ilocal, &
                  ' p:',i,j,k,' id:',index,dest,blocktype,request,cpl%desc_local(ilocal)%disp
#endif
            end if
            
          end do ! local couple arrays
          
        end do
      end do
    end do
    
    
    ! send the data
    n => rbtree_fd4_neigh_iterator_first(niter, cpl%neighstore)
    do while(associated(n))
      desttype => n%mpitype(FD4_NEIGH_CPL)
      if(desttype%count>0) then
        dest = niter%cur%key
        ! commit the datatype
        call fd4_mpitype_commit(desttype, err, startdispl=mindisp)
        FD4_L3_ASSERT(err/=0,'fd4_mpitype_commit failed')
        FD4_L3_ASSERT(.not.associated(mindisparray),'strange error')
#ifdef FD4_COUPLE_DEBUG
        do i=1,desttype%count
          write(300+mpi%rank,'(A,I5,I26,I26,I26)') &
              'TYPE ',desttype%types(i),desttype%indexes(i),desttype%displ(i),desttype%mindispl
        end do
        do i=1,desttype%count
          write(300+mpi%rank,'(A,I5,I26,I26,I26)') &
              'TYPE ',sortmdt%types(i),sortmdt%displ(i),sortmdt%sort(i)
        end do
#endif
        ! send the data, use count as tag
        call MPI_Isend(mindisparray, 1, desttype%datatype, dest, desttype%count, mpi%comm, request, err )
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Isend failed')
        call stack_push(mpi%requests, request)
      end if
      n => rbtree_fd4_neigh_iterator_next(niter)
    end do

        

    
    !
    !  2. receive data
    !


    ! iterate over ranks
    do dest=0,mpi%nprocs-1
    
      ! calculate intersection of rectangular unions of
      !  - all my local blocks (including offset due to face variables in couple context)
      !  - all of rank dest's couple arrays
      ! if this intersection is null, we can safely skip
      ! further partition matching checks of this rank
      if( .not. section_intersect(domain%blockunion, cpl%rankbnd(1:3,1:2,dest), cpl%faceoffset) ) cycle
      
      desttype => null()
    
      ! loop over couple arrays of rank dest
      do idesc=cpl%desc_rank(dest), cpl%desc_rank(dest+1) - 1
      
        ! if local array: calculate local index
        if(dest == mpi%rank) ilocal = idesc - cpl%desc_rank(dest) + 1

        ! set some convenience vars
        idx = cpl%desc_var%a(1,idesc)
        st  = cpl%desc_var%a(2,idesc)
        face = domain%vartab(idx)%facevar

        ! get bounds of the couple array within the spatial grid
        myabnd(1:3,1) = cpl%desc_abnd%a(1:3,cpl%desc_var%a(3,idesc)  )
        myabnd(1:3,2) = cpl%desc_abnd%a(1:3,cpl%desc_var%a(3,idesc)+1)

        ! stride for 4th dimension
        stride(1,1:2) = domain%varinfo%fields(idx,st,1:2)
        
        fext = 0
        if(face>0 .and. face<4) then
          fext(face) = fext(face) + 1
        end if      

        ! iterate over existing blocks and create an MPI type
        call fd4_iter_init(domain, iter)
        do while(associated(iter%cur))
          b => iter%cur

          bbnd = b%bnd

          if(face>0 .and. face<4) then
            ! additional cell plane if face var
            bbnd(face,2) = bbnd(face,2) + 1
          end if

          ! calculate intersection of block with array
          intersec(1:3,1)=MAX(bbnd(1:3,1),myabnd(1:3,1))
          intersec(1:3,2)=MIN(bbnd(1:3,2),myabnd(1:3,2))

#ifdef FD4_COUPLE_DEBUG
          case_str = '.'
#endif

          ! decide whether to send face cell or not depending on position of FD4 block vs. couple array
          if(cpl%face_overlap .and. face>0 .and. face<4) then
            ! CASE C: last cell of couple array >= last cell of block:
            ! transfer last cell, because right neighbor won't do that
#ifdef FD4_COUPLE_DEBUG
            case_str = 'C'
#endif
            if(myabnd(face,1)==bbnd(face,2)) then
              ! CASE B: first cell of couple array == last cell of block:
              ! do not transfer anything, left neighbor couple array will transfer all
              intersec(1,2) = intersec(1,1) - 1
#ifdef FD4_COUPLE_DEBUG
              case_str = 'B'
#endif
            else if(myabnd(face,2)<bbnd(face,2)) then
              ! CASE A: last cell of couple array is left of last cell of block:
              ! do not transfer last cell, will be transfered by right neighbor couple array
              intersec(face,2) = intersec(face,2) - 1
#ifdef FD4_COUPLE_DEBUG
              case_str = 'A'
#endif
            end if
          end if          

          ! check if intersection is not null
          if(intersec(1,1)<=intersec(1,2) .and. &
             intersec(2,1)<=intersec(2,2) .and. &
             intersec(3,1)<=intersec(3,2) ) then

            ! bounds of subset in block which respects to the intersection
            bbnd(1:3,1) = intersec(1:3,1) - b%bnd(1:3,1) + 1
            bbnd(1:3,2) = intersec(1:3,2) - b%bnd(1:3,1) + 1
             
            if(dest /= mpi%rank) then
              ! get MPI type for the block + variable
              ext = b%ext + fext
              call fd4_couple_blocktype(cpl%domain, idx, st, ext, bbnd, blocktype)
              call MPI_Get_address( b%sdata(face)%l(1,1,1,1), disp, err)
              ! index for sorting
              index = INT( fd4_domain_c2if(b%pos), i8k)   ! index for sorting
              if(.not. associated(desttype)) then
                ! check if a neigh type for this rank already exists, if not, create
                n => rbtree_fd4_neigh_get(cpl%neighstore, dest)
                if(.not. associated(n)) then
                  allocate(n, stat=err)
                  FD4_L3_ASSERT(err/=0,'allocate neigh')
                  call rbtree_fd4_neigh_insert (cpl%neighstore, dest, n)
                end if
                desttype => n%mpitype(FD4_NEIGH_FD4)
              end if
              call fd4_mpitype_add( desttype, index, 1, disp, blocktype, is_min)
              if(is_min) minptr => b%sdata(face)%l(1,1,1,1)
#ifdef FD4_COUPLE_DEBUG
              call MPI_Type_size(blocktype, request, err)
              write(300+mpi%rank,'(A,I5,I5,A5,A,I5,I5,I5,A,I10,I7,I12,I7,I26,A2)') &
                  'RECV ',idx,st,'',' p:',b%pos,' id:',index,dest,blocktype,request,disp,case_str
                  !,loc(b%sdata(face)%l(1,1,1,1))
#endif
            else
              !! copy local array section to block instead of sending msg to myself
              sbnd(1:3,1) = intersec(1:3,1) - myabnd(1:3,1) + cpl%desc_local(ilocal)%offset(1:3)
              sbnd(1:3,2) = intersec(1:3,2) - myabnd(1:3,1) + cpl%desc_local(ilocal)%offset(1:3)
              ! either array (nbins=1) or array4 (nbins>1) must be associated!
              if(domain%vartab(idx)%nbins == 1) then
                b%sdata(face)%l(stride(1,1),bbnd(1,1):bbnd(1,2),bbnd(2,1):bbnd(2,2),bbnd(3,1):bbnd(3,2)) &
                  = cpl%desc_local(ilocal)%array(sbnd(1,1):sbnd(1,2),sbnd(2,1):sbnd(2,2),sbnd(3,1):sbnd(3,2))
              else
                bnd4th(1) = cpl%desc_local(ilocal)%offset4
                bnd4th(2) = cpl%desc_local(ilocal)%offset4 + domain%vartab(idx)%nbins - 1
                b%sdata(face)%l(stride(1,1):stride(1,2),bbnd(1,1):bbnd(1,2),bbnd(2,1):bbnd(2,2),bbnd(3,1):bbnd(3,2)) &
                  = cpl%desc_local(ilocal)%array4(bnd4th(1):bnd4th(2),sbnd(1,1):sbnd(1,2),sbnd(2,1):sbnd(2,2),sbnd(3,1):sbnd(3,2))
              end if
#ifdef FD4_COUPLE_DEBUG
              write(300+mpi%rank,'(A,I5,I5,A5,A,3(I5),A2)') 'COPY ',idx,st,'',' p:',b%pos,case_str
#endif
            end if ! dest /= mpi%rank

          end if ! intersection

          call fd4_iter_next(iter)
        end do ! blocks
      
      end do ! desc

      ! now start receiving the data, use count as tag
      if(associated(desttype)) then
      if(desttype%count>0) then
        ! commit the datatype
        call fd4_mpitype_commit(desttype, err)
        FD4_L3_ASSERT(err/=0,'fd4_mpitype_commit failed')
        if(cpl%domain%stat_enabled) then
          call MPI_Type_size(desttype%datatype, i, err)
          cpl%domain%stat%values(cpl%stat_put) = cpl%domain%stat%values(cpl%stat_put) + i
        end if
#ifdef FD4_COUPLE_DEBUG
        do i=1,desttype%count
          write(300+mpi%rank,'(A,I5,I26,I26,I26)') &
              'TYPE ',desttype%types(i),desttype%indexes(i),desttype%displ(i),desttype%mindispl
        end do
        do i=1,desttype%count
          write(300+mpi%rank,'(A,I5,I26,I26,I26)') &
              'TYPE ',sortmdt%types(i),sortmdt%displ(i),sortmdt%sort(i)
        end do
#endif
#ifdef FD4_COUPLE_CHECK_RECV_SIZE
        ! blocking receive!!
        call MPI_Recv(minptr, 1, desttype%datatype, dest, desttype%count, mpi%comm, status, err)
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Recv failed')
        ! check if received msg size is same as expected msg size
        call MPI_Get_count(status, MPI_R_K, msgsize_recv, err)
        msgsize_recv = msgsize_recv * MPI_R_K_size
        call MPI_Type_size(desttype%datatype, msgsize_exp, err)
        if(msgsize_recv/=msgsize_exp) then
          write(fd4_out,'(A,A,I10,A,I10,A,I8)') FD4_LINEHD,'Error in fd4_couple_get: received: ',msgsize_recv, &
                                                           ' expected: ',msgsize_exp,' from rank',dest
          call MPI_Abort(MPI_COMM_WORLD,100,err)
        end if
#else
        ! recv the data
        call MPI_Irecv(minptr, 1, desttype%datatype, dest, desttype%count, mpi%comm, request, err)
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Irecv failed')
        call stack_push(mpi%requests, request)
#endif
      end if
      end if
      
    end do ! ranks


    !! wait for the communication to complete
    
    if(mpi%requests%pos>0) then
#ifdef FD4_COUPLE_DEBUG
       write(300+mpi%rank,*) 'WAIT ',mpi%requests%pos
#endif
      call MPI_Waitall(mpi%requests%pos, mpi%requests%a, MPI_STATUSES_IGNORE, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Waitall failed')
    end if    


    ! free MPI types, but remove neigh only, if the neigh has not been used this time
    dest = -1
    n => rbtree_fd4_neigh_iterator_first(niter, cpl%neighstore)
    do while(associated(n))
      if(n%mpitype(FD4_NEIGH_CPL)%count==0 .and. n%mpitype(FD4_NEIGH_FD4)%count==0) dest = niter%cur%key
      if(n%mpitype(FD4_NEIGH_CPL)%count>0) call fd4_mpitype_free(n%mpitype(FD4_NEIGH_CPL),.false.,.false.)
      if(n%mpitype(FD4_NEIGH_FD4)%count>0) call fd4_mpitype_free(n%mpitype(FD4_NEIGH_FD4),.false.,.false.)
      n => rbtree_fd4_neigh_iterator_next(niter)
      if(dest>=0) then
        ! delete previous item and fix iterator
        deleted => rbtree_fd4_neigh_delete(cpl%neighstore, dest, niter)
        call fd4_neigh_deallocate_ptr(deleted)
        dest = -1
      end if
    end do

    err = 0    
  
  end subroutine fd4_couple_put
  


  !! Get data from the framework. Each process asks for several blocks of
  !! data. $desc%abnd has to be set in all processes correctly. 
  !!
  !! *Additional limitation for face variables*: The upper face cell plane of the blocks is
  !! just cut and not transferred to the couple arrays!
  !!
  !! *TODO*:
  !!  - see limitations above
  !!  - much more parameter checking
  !!
  subroutine fd4_couple_get(cpl, err, opt_novnull)
    type(fd4_couple),     intent(inout),target :: cpl          !! couple context
    integer(i_k),         intent(out)          :: err          !! error status: 0...ok
    logical, optional,    intent(in)           :: opt_novnull  !! do not initalize the receiver arrays with vnull
    
    integer(i_k) :: ilocal, nbins
    type(fd4_mpi), pointer :: mpi
    integer(i8k) :: index
    integer(i_k) :: idx, st, idesc, lb(4)
    integer(i_k) :: i,j,k,bbnd(3,2),sbnd(3,2),intersec(3,2),face,stride(1,2),fext(3),ext(3),bnd4th(2)
    real(r_k), pointer :: minptr
    integer(i_k) :: myabnd(3,2)
    integer :: dest, blocktype, request
    integer(MPI_ADDRESS_KIND) :: disp
    type(fd4_block), pointer :: b
    type(fd4_iter), target  :: iter
    logical :: is_min
    type(fd4_domain), pointer :: domain
    real(r_k), pointer :: mindisparray             ! pointer to couple array with min. address
    integer(MPI_ADDRESS_KIND) :: mindisp           ! address of couple array with min. address
    type(fd4_mpitype), pointer :: desttype
    logical :: novnull
    integer(i_k) :: get_case, pos(3)
#if ( FD4_VERBOSE_LEVEL >= 3 )
    integer(i_k) :: realabnd(0:3,2)
#endif
#ifdef FD4_COUPLE_DEBUG
    character(1) :: case_str
#endif
#ifdef FD4_COUPLE_CHECK_RECV_SIZE
    integer :: status(MPI_STATUS_SIZE)
    integer :: msgsize_exp, msgsize_recv
#endif
    type(fd4_neigh), pointer :: n, deleted
    type(rbtree_fd4_neigh_iterator) :: niter

    err = -1
    FD4_L1_ASSERT(cpl%state/=CPL_CMMT,'cpl not commited')
    if(IAND(cpl%cpldir,FD4_CPL_GET)==0) then
      FD4_L1_ERROR('fd4_couple_get not possible')
    end if

    domain => cpl%domain
    mpi => domain%mpi
    minptr => NULL()
    nullify(mindisparray)
    mindisp = HUGE(1_MPI_ADDRESS_KIND)
    
    novnull = .false.
    if(present(opt_novnull)) novnull = opt_novnull


    ! loop over local couple arrays:
    ! check params and get array with minimum displacement
    ! and initialize coupling arrays with vnull
    do ilocal=1,cpl%ncplarrays
      idesc = ilocal + cpl%desc_rank(mpi%rank) - 1 ! index of this couple array in cpl%desc_var
      idx = cpl%desc_var%a(1,idesc)
      nbins = domain%vartab(idx)%nbins
      if( nbins == 1 .and. .not. associated(cpl%desc_local(ilocal)%array)) then
        FD4_L1_ERROR('desc_local%array not associated at '//i2s(ilocal))
      end if
      if( nbins >  1 .and. .not. associated(cpl%desc_local(ilocal)%array4)) then
        FD4_L1_ERROR('desc_local%array4 not associated at '//i2s(ilocal))
      end if
      ! find min displacement
      if( nbins > 1 ) then
        lb(1:4) = lbound(cpl%desc_local(ilocal)%array4)
        call MPI_Get_address( cpl%desc_local(ilocal)%array4(lb(1),lb(2),lb(3),lb(4)), disp, err)
        if(disp < mindisp) then
          mindisp = disp
          mindisparray => cpl%desc_local(ilocal)%array4(lb(1),lb(2),lb(3),lb(4))
        end if
      else
        lb(1:3) = lbound(cpl%desc_local(ilocal)%array)
        call MPI_Get_address( cpl%desc_local(ilocal)%array(lb(1),lb(2),lb(3)), disp, err)
        if(disp < mindisp) then
          mindisp = disp
          mindisparray => cpl%desc_local(ilocal)%array(lb(1),lb(2),lb(3))
        end if
      end if
      ! save array address
      cpl%desc_local(ilocal)%disp = disp
      ! get bounds of the couple array within the spatial grid
      myabnd(1:3,1) = cpl%desc_abnd%a(1:3,cpl%desc_var%a(3,idesc)  )
      myabnd(1:3,2) = cpl%desc_abnd%a(1:3,cpl%desc_var%a(3,idesc)+1)
#if ( FD4_VERBOSE_LEVEL >= 3 )
      ! get real/allocated bounds of the couple array
      if( nbins > 1 ) then
        realabnd(0:3,1) = lbound(cpl%desc_local(ilocal)%array4)
        realabnd(0:3,2) = ubound(cpl%desc_local(ilocal)%array4)
      else
        realabnd(0,1:2) = 0
        realabnd(1:3,1) = lbound(cpl%desc_local(ilocal)%array)
        realabnd(1:3,2) = ubound(cpl%desc_local(ilocal)%array)
      end if
      ! check array size for 4th dimension
      if( nbins > 1 ) then
        if( realabnd(0,1) > cpl%desc_local(ilocal)%offset4 ) then
          FD4_L1_ERROR('desc_local%offset4 too small at '//i2s(ilocal))
        end if
        if( realabnd(0,2) < cpl%desc_local(ilocal)%offset4 + nbins - 1 ) then
          FD4_L1_ERROR('desc_local%array4 too small at '//i2s(ilocal))
        end if
      end if      
      ! check array size for spatial dimensions      
      if( ANY( realabnd(1:3,1) > cpl%desc_local(ilocal)%offset ) ) then
        FD4_L1_ERROR('desc%offset too small at '//i2s(idesc))
      end if
      if( ANY( realabnd(1:3,2) < cpl%desc_local(ilocal)%offset + myabnd(1:3,2) - myabnd(1:3,1) ) ) then
        FD4_L1_ERROR('desc%array too small at '//i2s(idesc))
      end if
#endif
      ! initialize array with vnull for specified variable
      if(.not. novnull) then
        sbnd(1:3,1) = cpl%desc_local(ilocal)%offset
        sbnd(1:3,2) = cpl%desc_local(ilocal)%offset + myabnd(1:3,2) - myabnd(1:3,1)
        if( nbins > 1 ) then
          bnd4th(1) = cpl%desc_local(ilocal)%offset4
          bnd4th(2) = cpl%desc_local(ilocal)%offset4 + nbins - 1
          cpl%desc_local(ilocal)%array4(bnd4th(1):bnd4th(2),sbnd(1,1):sbnd(1,2),sbnd(2,1):sbnd(2,2),sbnd(3,1):sbnd(3,2)) &
            = domain%vartab(idx)%vnull
        else
          cpl%desc_local(ilocal)%array(sbnd(1,1):sbnd(1,2),sbnd(2,1):sbnd(2,2),sbnd(3,1):sbnd(3,2)) & 
            = domain%vartab(idx)%vnull
        end if
      end if
    end do

    mpi%requests%pos = 0
 
#ifdef FD4_COUPLE_DEBUG
    write(300+mpi%rank,'(A,I5)')  '---- fd4_couple_get ',cpl%ncplarrays
#endif

    
    !
    !  1. send data
    !


    ! iterate over ranks
    do dest=0,mpi%nprocs-1

      ! calculate intersection of rectangular unions of
      !  - all my local blocks (including offset due to face variables in couple context)
      !  - all of rank dest's couple arrays
      ! if this intersection is null, we can safely skip
      ! further partition matching checks of this rank
      if( .not. section_intersect(domain%blockunion, cpl%rankbnd(1:3,1:2,dest), cpl%faceoffset) ) cycle

      desttype => null()

      ! loop over couple arrays of rank dest
      do idesc=cpl%desc_rank(dest), cpl%desc_rank(dest+1) - 1

        ! if local array: calculate local index
        if(dest == mpi%rank) ilocal = idesc - cpl%desc_rank(dest) + 1

        ! set some convenience vars
        idx = cpl%desc_var%a(1,idesc)
        st  = cpl%desc_var%a(2,idesc)
        face = domain%vartab(idx)%facevar

        ! get bounds of the couple array within the spatial grid
        myabnd(1:3,1) = cpl%desc_abnd%a(1:3,cpl%desc_var%a(3,idesc)  )
        myabnd(1:3,2) = cpl%desc_abnd%a(1:3,cpl%desc_var%a(3,idesc)+1)

        ! stride for 4th dimension
        stride(1,1:2) = domain%varinfo%fields(idx,st,1:2)
        
        fext = 0
        if(face>0 .and. face<4) then
          fext(face) = fext(face) + 1
        end if      

        ! iterate over existing blocks and create an MPI type
        call fd4_iter_init(domain, iter)
        do while(associated(iter%cur))
          b => iter%cur

          bbnd = b%bnd

          ! additional cell row if face variable
          if(face>0 .and. face<4) bbnd(face,2) = bbnd(face,2) + 1

          ! calculate intersection of block with array
          intersec(1:3,1)=MAX(bbnd(1:3,1),myabnd(1:3,1))
          intersec(1:3,2)=MIN(bbnd(1:3,2),myabnd(1:3,2))

#ifdef FD4_COUPLE_DEBUG
          case_str = '.'
#endif
          ! face variable: we need some extra checks to determine what to send
          if(face>0 .and. face<4) then
#ifdef FD4_COUPLE_DEBUG
            case_str = 'A'
#endif
            pos(:) = b%pos(:)
            if( myabnd(face,2) == bbnd(face,1) ) then
              ! case B: last cell of couple array is first cell of FD4 block
              pos(face) = pos(face) - 1
#ifdef FD4_COUPLE_DEBUG
              case_str = 'b'
#endif
              if( domain%blkmap(pos(1),pos(2),pos(3),domain%blkmap_now) >= 0 ) then
                ! left neighbor present, so we don't need to send data
                intersec(1,2) = intersec(1,1) - 1  ! set intersection to zero
              end if
            else if( myabnd(face,2) > bbnd(face,2) ) then
              ! case C: last cell of couple array is right outside the FD4 block
              pos(face) = pos(face) + 1
#ifdef FD4_COUPLE_DEBUG
              case_str = 'c'
#endif
              if( domain%blkmap(pos(1),pos(2),pos(3),domain%blkmap_now) >= 0 ) then
                ! right neighbor provides additional face cell
                intersec(face,2) = intersec(face,2) - 1
#ifdef FD4_COUPLE_DEBUG
                case_str = 'C'
#endif
              end if
            end if
          end if

          ! check if intersection is not null
          if(intersec(1,1)<=intersec(1,2) .and. &
             intersec(2,1)<=intersec(2,2) .and. &
             intersec(3,1)<=intersec(3,2) ) then

            ! bounds of subset in block which respects to the intersection
            bbnd(1:3,1) = intersec(1:3,1) - b%bnd(1:3,1) + 1
            bbnd(1:3,2) = intersec(1:3,2) - b%bnd(1:3,1) + 1
             
            if(dest /= mpi%rank) then
              ! get MPI type for the block + variable
              ext = b%ext + fext
              call fd4_couple_blocktype(cpl%domain, idx, st, ext, bbnd, blocktype)
              call MPI_Get_address( b%sdata(face)%l(1,1,1,1), disp, err)
              ! index for sorting
              index = INT( fd4_domain_c2if(b%pos), i8k)   ! index for sorting
              if(.not. associated(desttype)) then
                ! check if a neigh type for this rank already exists, if not, create
                n => rbtree_fd4_neigh_get(cpl%neighstore, dest)
                if(.not. associated(n)) then
                  allocate(n, stat=err)
                  FD4_L3_ASSERT(err/=0,'allocate neigh')
                  call rbtree_fd4_neigh_insert (cpl%neighstore, dest, n)
                end if
                desttype => n%mpitype(FD4_NEIGH_FD4)
              end if
              call fd4_mpitype_add( desttype, index, 1, disp, blocktype, is_min)
              if(is_min) minptr => b%sdata(face)%l(1,1,1,1)
#ifdef FD4_COUPLE_DEBUG
              call MPI_Type_size(blocktype, request, err)
              write(300+mpi%rank,'(A,I5,I5,A5,A,I5,I5,I5,A,I10,I7,I12,I7,A2)') &
                  'SEND ',idx,st,'',' p:',b%pos,' id:',index,dest,blocktype,request,case_str
#endif
            else
              ! copy local array section to couple array instead of sending msg to myself
              sbnd(1:3,1) = intersec(1:3,1) - myabnd(1:3,1) + cpl%desc_local(ilocal)%offset(1:3)
              sbnd(1:3,2) = intersec(1:3,2) - myabnd(1:3,1) + cpl%desc_local(ilocal)%offset(1:3)
              ! either array (nbins=1) or array4 (nbins>1) must be associated!
              if(domain%vartab(idx)%nbins == 1) then
                cpl%desc_local(ilocal)%array(sbnd(1,1):sbnd(1,2),sbnd(2,1):sbnd(2,2),sbnd(3,1):sbnd(3,2)) &
                  = b%sdata(face)%l(stride(1,1),bbnd(1,1):bbnd(1,2),bbnd(2,1):bbnd(2,2),bbnd(3,1):bbnd(3,2))
              else
                bnd4th(1) = cpl%desc_local(ilocal)%offset4
                bnd4th(2) = cpl%desc_local(ilocal)%offset4 + domain%vartab(idx)%nbins - 1
                cpl%desc_local(ilocal)%array4(bnd4th(1):bnd4th(2),sbnd(1,1):sbnd(1,2),sbnd(2,1):sbnd(2,2),sbnd(3,1):sbnd(3,2)) &
                  = b%sdata(face)%l(stride(1,1):stride(1,2),bbnd(1,1):bbnd(1,2),bbnd(2,1):bbnd(2,2),bbnd(3,1):bbnd(3,2))
                
              end if
#ifdef FD4_COUPLE_DEBUG
              write(300+mpi%rank,'(A,I5,I5,A5,A,3(I5),A2)') 'COPY ',idx,st,'',' p:',b%pos,case_str
#endif
            end if ! dest /= mpi%rank

          end if ! intersection

          call fd4_iter_next(iter)
        end do ! blocks


      
      end do ! desc

      
      ! now start sending the data, use count as tag
      if(associated(desttype)) then
      if(desttype%count>0) then
        ! commit the datatype
        call fd4_mpitype_commit(desttype, err)
        FD4_L3_ASSERT(err/=0,'fd4_mpitype_commit failed')
        if(cpl%domain%stat_enabled) then
          call MPI_Type_size(desttype%datatype, i, err)
          cpl%domain%stat%values(cpl%stat_get) = cpl%domain%stat%values(cpl%stat_get) + i
        end if
        ! recv the data
        call MPI_Isend(minptr, 1, desttype%datatype, dest, desttype%count, mpi%comm, request, err )
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Isend failed')
        call stack_push(mpi%requests, request)
      end if
      end if
      

    end do ! ranks


    
    !
    !  2. receive data
    !

    ! loop over all blocks of the domain
    ! to build an MPI datatype to recv subsections
    ! from the blocks of other processes to my couple arrays
    do k=cpl%loopbnd(3,1),cpl%loopbnd(3,2)
      do j=cpl%loopbnd(2,1),cpl%loopbnd(2,2)
        do i=cpl%loopbnd(1,1),cpl%loopbnd(1,2)
        
          ! cycle the loop if block does not exist
          dest = domain%blkmap(i,j,k,domain%blkmap_now)
          if(dest == -1 ) cycle

          ! cycle if this is my block
          if(dest == mpi%rank) cycle

          ! loop over local couple arrays
          do ilocal=1,cpl%ncplarrays
            get_case = cpl%arraytypes(ilocal, i, j, k, CPL_GET_CASE)
            if(get_case == CPL_GET_CASE_SKIP) cycle

            ! index of this couple array in cpl%desc_var
            idesc = ilocal + cpl%desc_rank(mpi%rank) - 1
            
            ! get the matching type from the array of pre-built datatypes
            blocktype = cpl%arraytypes(ilocal, i, j, k, CPL_GET)
#ifdef FD4_COUPLE_DEBUG
            case_str = '.'
#endif
            
            if(get_case /= CPL_GET_CASE_NORMAL) then
              ! we need to check face-neighbor allocation status to decide what to transfer
              face = domain%vartab( cpl%desc_var%a(1,idesc) )%facevar
              pos(1) = i; pos(2) = j; pos(3) = k
#ifdef FD4_COUPLE_DEBUG
            case_str = 'A'
#endif
              ! depending on the case (position of couple array vs. FD4 block and neighbor allocation status)
              ! we might skip transfer of this subsection or use different mpi type
              if(get_case == CPL_GET_CASE_CHECK_L) then
                ! case B
                pos(face) = pos(face) - 1
#ifdef FD4_COUPLE_DEBUG
                case_str = 'b'
#endif
                if( domain%blkmap(pos(1),pos(2),pos(3),domain%blkmap_now) >= 0 ) then
                  ! left neighbor FD4 block present: do not transfer anything
                  blocktype = MPI_DATATYPE_NULL
                end if
              else if(get_case == CPL_GET_CASE_CHECK_R) then
                ! case C
                pos(face) = pos(face) + 1
#ifdef FD4_COUPLE_DEBUG
                case_str = 'c'
#endif
                if( domain%blkmap(pos(1),pos(2),pos(3),domain%blkmap_now) >= 0 ) then
                  ! right neigbor FD4 block present: use type without additional face cell
                  blocktype = cpl%arraytypes(ilocal, i, j, k, CPL_GET_NOFACE)
#ifdef FD4_COUPLE_DEBUG
                  case_str = 'C'
#endif
                end if
              end if
              
            end if
            
            ! check if type exists
            if(blocktype /= MPI_DATATYPE_NULL) then

              index = INT( fd4_domain_ijk2if(i,j,k), i8k)   ! index for sorting

              ! check if a neigh type for this rank already exists, if not, create
              n => rbtree_fd4_neigh_get(cpl%neighstore, dest)
              if(.not. associated(n)) then
                allocate(n, stat=err)
                FD4_L3_ASSERT(err/=0,'allocate neigh')
                call rbtree_fd4_neigh_insert (cpl%neighstore, dest, n)
              end if

              call fd4_mpitype_add(n%mpitype(FD4_NEIGH_CPL), index, 1, cpl%desc_local(ilocal)%disp, blocktype)
#ifdef FD4_COUPLE_DEBUG
              call MPI_Type_size(blocktype, request, err)
              write(300+mpi%rank,'(A,I5,I5,I5,A,I5,I5,I5,A,I10,I7,I12,I7,A2)')       &
                  'RECV ',cpl%desc_var%a(1,idesc),cpl%desc_var%a(2,idesc),ilocal, &
                  ' p:',i,j,k,' id:',index,dest,blocktype,request,case_str
#endif
            end if

          end do ! local couple arrays
          
        end do
      end do
    end do
    
    
    ! recv the data
    n => rbtree_fd4_neigh_iterator_first(niter, cpl%neighstore)
    do while(associated(n))
      desttype => n%mpitype(FD4_NEIGH_CPL)
      if(desttype%count>0) then
        dest = niter%cur%key
        ! commit the datatype
        call fd4_mpitype_commit(desttype, err, startdispl=mindisp)
        FD4_L3_ASSERT(err/=0,'fd4_mpitype_commit failed')
        FD4_L3_ASSERT(.not.associated(mindisparray),'strange error')
#ifdef FD4_COUPLE_CHECK_RECV_SIZE
        ! blocking receive!!
        call MPI_Recv(mindisparray, 1, desttype%datatype, dest, desttype%count, mpi%comm, status, err )
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Recv failed')
        ! check if received msg size is same as expected msg size
        call MPI_Get_count(status, MPI_R_K, msgsize_recv, err)
        msgsize_recv = msgsize_recv * MPI_R_K_size
        call MPI_Type_size(desttype%datatype, msgsize_exp, err)
        if(msgsize_recv/=msgsize_exp) then
          write(fd4_out,'(A,A,I10,A,I10,A,I8)') FD4_LINEHD,'Error in fd4_couple_get: received: ',msgsize_recv, &
                                                           ' expected: ',msgsize_exp,' from rank',dest
          call MPI_Abort(MPI_COMM_WORLD,200,err)
        end if
#else
        ! recv  the data, use count as tag
        call MPI_Irecv(mindisparray, 1, desttype%datatype, dest, desttype%count, mpi%comm, request, err )
        err = err - MPI_SUCCESS
        FD4_L3_ASSERT(err/=0,'MPI_Irecv failed')
        call stack_push(mpi%requests, request)
#endif
      end if
      n => rbtree_fd4_neigh_iterator_next(niter)
    end do      


    !! wait for the communication to complete
    
    if(mpi%requests%pos>0) then
#ifdef FD4_COUPLE_DEBUG
       write(300+mpi%rank,*) 'WAIT ',mpi%requests%pos
#endif
      call MPI_Waitall(mpi%requests%pos, mpi%requests%a, MPI_STATUSES_IGNORE, err)
      err = err - MPI_SUCCESS
      FD4_L3_ASSERT(err/=0,'MPI_Waitall failed')
    end if    


    ! free MPI types, but remove neigh only, if the neigh has not been used this time
    dest = -1
    n => rbtree_fd4_neigh_iterator_first(niter, cpl%neighstore)
    do while(associated(n))
      if(n%mpitype(FD4_NEIGH_CPL)%count==0 .and. n%mpitype(FD4_NEIGH_FD4)%count==0) dest = niter%cur%key
      if(n%mpitype(FD4_NEIGH_CPL)%count>0) call fd4_mpitype_free(n%mpitype(FD4_NEIGH_CPL),.false.,.false.)
      if(n%mpitype(FD4_NEIGH_FD4)%count>0) call fd4_mpitype_free(n%mpitype(FD4_NEIGH_FD4),.false.,.false.)
      n => rbtree_fd4_neigh_iterator_next(niter)
      if(dest>=0) then
        ! delete previous item and fix iterator
        deleted => rbtree_fd4_neigh_delete(cpl%neighstore, dest, niter)
        call fd4_neigh_deallocate_ptr(deleted)
        dest = -1
      end if
    end do

    err = 0    
  
  end subroutine fd4_couple_get





  !! Get an MPI data type for the variable described by ($idx, $st), a block
  !! of size $ext, and a spatial subsection within the block described
  !! by $bbnd.
  !! If an MPI data type with same properties has already been created,
  !! this type will be re-used.
  !! 
  !! Internal routine. Used in $fd4_couple_put and $fd4_couple_get.
  subroutine fd4_couple_blocktype(domain, idx, st, ext, bnd, blocktype)
    type(fd4_domain), intent(inout)  :: domain  !! the domain
    integer(i_k), intent(in)  :: idx            !! variable index
    integer(i_k), intent(in)  :: st             !! time step index
    integer(i_k), intent(in)  :: ext(3)         !! block extents
    integer(i_k), intent(in)  :: bnd(3,2)       !! bound of spatial subset
    integer,      intent(out) :: blocktype      !! MPI data type for the block
    
    integer(i8k) :: index
    integer(i_k) :: err, stride(1,2), ext4(0:3)
    logical :: blocktype_present
    
    ! get index of the type in domain-global cpltypestore
    call fd4_domain_block_encode(domain%cplblkenc, idx, st, ext(1:3), bnd, index)

    ! check if the type is present already
    blocktype = rbtree_int8_get(domain%cpltypestore, index, found = blocktype_present)

#ifdef FD4_COUPLE_DEBUG
    !write(300+domain%mpi%rank,'(A,I5,I5,A5,A,I5,I5,I5,A,3(I5),A,3(I5),L5,I16)') &
    !    'BLTY ',idx,st,'',' ext:',ext,'  bnd:',bnd(:,1),'-',bnd(:,2), blocktype_present, index
#endif

    if(.not. blocktype_present) then
      stride(1,1:2) = domain%varinfo%fields(idx,st,1:2)
      ext4(0) = domain%varinfo%field_len(domain%vartab(idx)%facevar)
      ext4(1:3) = ext(1:3)
      !VT_USER_START('FD4_COUPLE_PUT.build_array_type')
      call fd4_comm_build_array_type(ext4, stride, bnd, blocktype, err)
      !VT_USER_END('FD4_COUPLE_PUT.build_array_type')
      FD4_L3_ASSERT(err/=0,'fd4_comm_build_array_type failed')
      ! put type to blockstore
      call rbtree_int8_insert(domain%cpltypestore, index, blocktype, err)
      FD4_L3_ASSERT(err/=0,'rbtree_int_insert failed')
    end if

  end subroutine fd4_couple_blocktype



  !! Pretend putting data to the framework and mark all framework blocks
  !! which need to be created to fully capture all threshold variables.
  !!
  !! Use this in a 3-stage coupling process:
  !!  - $fd4_couple_mark_blocks to mark required blocks
  !!  - $fd4_balance_readjust to ensure these blocks are created and balanced
  !!  - $fd4_couple_put to finally put the data to fd4
  !! You must use the same parameters for $fd4_couple_mark_blocks as for $fd4_couple_put.
  !!
  !! *TODO*:
  !!  - explicitly loop over all grid cells of a block instead using any(abs(block))
  !!  - see limitations above
  !!  - much more parameter checking
  !!
  subroutine fd4_couple_mark_blocks(cpl, err)
    type(fd4_couple),     intent(inout),target :: cpl          !! couple context
    integer(i_k),         intent(out)          :: err          !! error status: 0...ok
    
    integer(i_k) :: ilocal, nbins
    type(fd4_mpi), pointer :: mpi
    integer(i_k) :: idx, idesc
    integer(i_k) :: i,j,k,bbnd(3,2),sbnd(3,2),intersec(3,2),face,bnd4th(2)
    integer(i_k) :: myabnd(3,2)
    type(fd4_domain), pointer :: domain
#if ( FD4_VERBOSE_LEVEL >= 3 )
    integer(i_k) :: realabnd(0:3,2)
#endif

    err = -1
    FD4_L1_ASSERT(cpl%state/=CPL_CMMT,'cpl not commited')
    if(IAND(cpl%cpldir,FD4_CPL_PUT)==0) then
      FD4_L1_ERROR('fd4_couple_mark_blocks not possible')
    end if
   
    domain => cpl%domain
    mpi => domain%mpi
    intersec(:,1) = 0
    intersec(:,2) = -1
    
    ! loop over local couple arrays:
    ! check params and get array with minimum displacement
    do ilocal=1,cpl%ncplarrays
      idesc = ilocal + cpl%desc_rank(mpi%rank) - 1 ! index of this couple array in cpl%desc_var
      idx = cpl%desc_var%a(1,idesc)
      nbins = domain%vartab(idx)%nbins
      if( nbins == 1 .and. .not. associated(cpl%desc_local(ilocal)%array)) then
        FD4_L1_ERROR('desc_local%array not associated at '//i2s(ilocal))
      end if
      if( nbins >  1 .and. .not. associated(cpl%desc_local(ilocal)%array4)) then
        FD4_L1_ERROR('desc_local%array4 not associated at '//i2s(ilocal))
      end if
#if ( FD4_VERBOSE_LEVEL >= 3 )
      ! get real/allocated bounds of the couple array
      if( nbins > 1 ) then
        realabnd(0:3,1) = lbound(cpl%desc_local(ilocal)%array4)
        realabnd(0:3,2) = ubound(cpl%desc_local(ilocal)%array4)
      else
        realabnd(0,1:2) = 0
        realabnd(1:3,1) = lbound(cpl%desc_local(ilocal)%array)
        realabnd(1:3,2) = ubound(cpl%desc_local(ilocal)%array)
      end if
      ! check array size for 4th dimension
      if( nbins > 1 ) then
        if( realabnd(0,1) > cpl%desc_local(ilocal)%offset4 ) then
          FD4_L1_ERROR('desc_local%offset4 too small at '//i2s(ilocal))
        end if
        if( realabnd(0,2) < cpl%desc_local(ilocal)%offset4 + nbins - 1 ) then
          FD4_L1_ERROR('desc_local%array4 too small at '//i2s(ilocal))
        end if
      end if      
      ! get bounds of the couple array within the spatial grid
      myabnd(1:3,1) = cpl%desc_abnd%a(1:3,cpl%desc_var%a(3,idesc)  )
      myabnd(1:3,2) = cpl%desc_abnd%a(1:3,cpl%desc_var%a(3,idesc)+1)
      ! check array size for spatial dimensions      
      if( ANY( realabnd(1:3,1) > cpl%desc_local(ilocal)%offset ) ) then
        FD4_L1_ERROR('desc%offset too small at '//i2s(idesc))
      end if
      if( ANY( realabnd(1:3,2) < cpl%desc_local(ilocal)%offset + myabnd(1:3,2) - myabnd(1:3,1) ) ) then
        FD4_L1_ERROR('desc%array too small at '//i2s(idesc))
      end if
#endif
    end do

    
    ! loop over all blocks of the domain

    do k=cpl%loopbnd(3,1),cpl%loopbnd(3,2)
      do j=cpl%loopbnd(2,1),cpl%loopbnd(2,2)
        do i=cpl%loopbnd(1,1),cpl%loopbnd(1,2)
        
          ! cycle the loop if block does exist
          !    actually, this is a bit inaccurate!! fd4_balance_readjust
          !    may decide to delete a block which we may found to be
          !    required - if we would not cycle now. but running
          !    this check for all blocks is extremely costly for high
          !    block numbers.
          !! if(domain%blkmap(i,j,k,domain%blkmap_now) > -1 ) cycle

          ! loop over local couple arrays
          local_loop: do ilocal=1,cpl%ncplarrays
            idesc = ilocal + cpl%desc_rank(mpi%rank) - 1 ! index of this couple array in cpl%desc_var

            idx = cpl%desc_var%a(1,idesc)

            ! cycle the loop if this var has no threshold
            if(domain%vartab(idx)%vthres == FD4_NOTHRES) cycle

            ! set some convenience vars
            face = domain%vartab(idx)%facevar

            ! get bounds of the couple array within the spatial grid
            myabnd(1:3,1) = cpl%desc_abnd%a(1:3,cpl%desc_var%a(3,idesc)  )
            myabnd(1:3,2) = cpl%desc_abnd%a(1:3,cpl%desc_var%a(3,idesc)+1)
            
            ! bounds of this non-existing block
            bbnd(1,:) = (/ domain%d(1)%bl(i) , domain%d(1)%bl(i+1) - 1 /)
            bbnd(2,:) = (/ domain%d(2)%bl(j) , domain%d(2)%bl(j+1) - 1 /)
            bbnd(3,:) = (/ domain%d(3)%bl(k) , domain%d(3)%bl(k+1) - 1 /)

            if(face>0 .and. face<4) then
              ! additional cell plane if face var
              bbnd(face,2) = bbnd(face,2) + 1
            end if

            ! calculate intersection of block with array
            intersec(1:3,1)=MAX(bbnd(1:3,1),myabnd(1:3,1))
            intersec(1:3,2)=MIN(bbnd(1:3,2),myabnd(1:3,2))

            ! check if intersection is not null
            if(intersec(1,1)<=intersec(1,2) .and. &
               intersec(2,1)<=intersec(2,2) .and. &
               intersec(3,1)<=intersec(3,2) ) then
              ! bounds of subset in array which respects to the current block, array content starts at given offset
              sbnd(1:3,1) = intersec(1:3,1) - myabnd(1:3,1) + cpl%desc_local(ilocal)%offset(1:3)
              sbnd(1:3,2) = intersec(1:3,2) - myabnd(1:3,1) + cpl%desc_local(ilocal)%offset(1:3)
              ! either array (nbins=1) or array4 (nbins>1) must be associated!
              if(domain%vartab(idx)%nbins == 1) then
                ! if any of the values in the intersection is larger than vthres, we mark this block
                !TODO: any(abs(...)) may be quiet expensive, explicitly loop over grid cells instead!
                if(   ANY( abs(cpl%desc_local(ilocal)%array( &
                   sbnd(1,1):sbnd(1,2),sbnd(2,1):sbnd(2,2),sbnd(3,1):sbnd(3,2) ) ) &
                    > domain%vartab(idx)%vthres ) ) then
                  call stack_push( domain%markedblocks, fd4_domain_ijk2if(i,j,k) )
                  ! we marked this block, so we need not to evaluate the other descriptions at this pos
                  exit local_loop
                end if
              else
                ! if any of the values in the intersection is larger than vthres, we mark this block
                bnd4th(1) = cpl%desc_local(ilocal)%offset4
                bnd4th(2) = cpl%desc_local(ilocal)%offset4 + domain%vartab(idx)%nbins - 1
                if(   ANY( abs(cpl%desc_local(ilocal)%array4( &
                   bnd4th(1):bnd4th(2),sbnd(1,1):sbnd(1,2),sbnd(2,1):sbnd(2,2),sbnd(3,1):sbnd(3,2) ) ) &
                    > domain%vartab(idx)%vthres ) ) then
                  call stack_push( domain%markedblocks, fd4_domain_ijk2if(i,j,k) )
                  ! we marked this block, so we need not to evaluate the other descriptions at this pos
                  exit local_loop
                end if
              end if

            end if ! intersection

          end do local_loop ! local couple arrays
          
        end do
      end do
    end do
    
    ! merge the local stacks to one global
    call fd4_couple_stack_global_merge(domain%mpi, domain%markedblocks, err)
    FD4_L3_ASSERT(err/=0,'fd4_couple_stack_global_merge failed')


    ! important: set the flag, that we have marked blocks,
    !            so fd4_balance_readjust knows this
    domain%blocks_are_marked = ( domain%markedblocks%pos > 0 )
    
    err = 0    

  end subroutine fd4_couple_mark_blocks




  !! Merge the process-local stacks $st to one global stack in every
  !! process.
  !!
  !! Duplicates will be removed, the stack will be sorted.
  !! It is assumed that the process local stacks are already sorted.
  !!
  !! Internal routine.
  !!
  subroutine fd4_couple_stack_global_merge(mpi, st, err)
    type(fd4_mpi),     intent(inout) :: mpi               !! the domain's mpi context
    type(stack),       intent(inout) :: st                !! the stack 
    integer(i_k),      intent(out)   :: err               !! error status: 0...ok
    
    integer :: count, i
    type(stack) :: st_global
    type(rbtree_int) :: intree
    type(rbtree_int_iterator) :: initer
    integer(i_k) :: p, stat
    type(stack) :: st_sort

    err = -1   

    !do i=1,st%pos
    !  write(300+mpi%rank,'(I20,I5,I5,I5)') st%a(i),fd4_domain_i2cf(st%a(i))
    !end do
    
    !! tell all other processes how many entries my local stack3 has
    call MPI_Allgather( st%pos, 1, MPI_I_K, mpi%recvcnt, 1, MPI_I_K, mpi%comm, err )
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Allgather failed')
    
    !! let's see if there's anything, if no, we can leave this routine
    count = SUM(mpi%recvcnt)
    if(count == 0) then
      return
    end if
        
    !! now collect the contents of the local stacks
    
    !! set receive displacements for MPI_Allgatherv
    mpi%recvdis(0) = 0
    do i=0,mpi%nprocs-2
      mpi%recvdis(i+1) = mpi%recvdis(i) + mpi%recvcnt(i)
    end do

    !! allocate memory in the global stack
    call stack_create(st_global, count)
    st_global%pos = count

    !! go! receive directly into the new created st_global
    call MPI_Allgatherv (       st%a,       st%pos,                MPI_I_K, &
                         st_global%a,  mpi%recvcnt,   mpi%recvdis, MPI_I_K, mpi%comm, err )
    err = err - MPI_SUCCESS
    FD4_L3_ASSERT(err/=0,'MPI_Allgatherv failed')
    
    call stack_move(st_global, st)

    !! sort, i.e. merge the sorted parts of the st_global to one sorted stack st_sort

    call stack_create(st_sort, count)
    ! senddis contains the end indices of the procs in st
    mpi%senddis(:) = mpi%recvdis(:) + mpi%recvcnt(:)
    ! recvdis contains the start indices of the procs in st (see above)
    mpi%recvdis(:) = mpi%recvdis(:) + 1
    
    ! initialize the tree of integers (maps block idx to rank, sorting by block idx)
    !  holds first block idx of each rank, automatic sorting to find
    !  smallest block idx
    do p=0,mpi%nprocs-1
      if(mpi%recvdis(p)>mpi%senddis(p)) cycle
      do
        call rbtree_int_insert(intree, st%a( mpi%recvdis(p) ), p, stat )
        if(stat==0) exit
        mpi%recvdis(p) = mpi%recvdis(p) + 1
        if(mpi%recvdis(p)>mpi%senddis(p)) exit
      end do
    end do
    
    !call rbtree_int_print(intree)
    
    do while(intree%length>0)
      ! first element = smallest remaining block index 
      p = rbtree_int_iterator_first(initer,intree)
      ! push this value to the sorted stack
      st_sort%pos = st_sort%pos + 1
      !print *,st_sort%pos,'of',size(st_sort%a),p,initer%cur%key
      FD4_L3_ASSERT(st_sort%pos>size(st_sort%a),'st_sort too small!')
      FD4_L3_ASSERT(initer%cur%key/=st%a(mpi%recvdis(p)),'tree or stack corrupt!')
      st_sort%a(st_sort%pos) = initer%cur%key
      ! delete this element, this invalidates the iterator
      p = rbtree_int_delete(intree, initer%cur%key)
      ! increase start index
      mpi%recvdis(p) = mpi%recvdis(p) + 1
      ! put the min block idx of this rank to the tree
      if(mpi%recvdis(p)>mpi%senddis(p)) cycle
      do
        call rbtree_int_insert(intree, st%a( mpi%recvdis(p) ), p, stat )
        if(stat==0) exit
        ! need to take next item, because of duplicates
        mpi%recvdis(p) = mpi%recvdis(p) + 1
        if(mpi%recvdis(p)>mpi%senddis(p)) exit
      end do
    end do

    ! this is actually not necessary
    call rbtree_int_clear(intree)

    call stack_move(st_sort, st)
   
    !do i=1,st%pos
    !  write(200+mpi%rank,'(I20,I5,I5,I5)') st%a(i),fd4_domain_i2cf(st%a(i))
    !end do
    
    err = 0

  end subroutine fd4_couple_stack_global_merge
  

end module fd4_couple_mod
